"""
Ploomber execution model:

To orchestrate Task execution, they are organized in a DAG object that keeps
track of dependencies and status.

It all starts with a call to DAG.render(), all placeholders are resolved
and the task status is determined. If any task fails to render, the process
is stopped and such DAG cannot be executed.

Once all tasks render successfully, their product's metadata determined
whether they should run or not based on source code changes, this behavior
can be overridden by passing DAG.render(force=True).

When DAG.build() is called, actual execution happens. Tasks are executed in
order (but up-to-date tasks are skipped). If tasks succeed, their downstream
dependencies will be executed, otherwise they are aborted.

If any task fails, DAG executes the on_failure hook, otherwise, it executes
the on_finish hook.

The DAG does not execute tasks, but delegates this to an Executor object,
executors adhere to task status and do not build tasks if they are marked
as Aborted or Skipped. But executors are free to execute tasks in
subprocesses, in such case, the have to report back the result of task
building to the main process.

DAGs can be executed in different forms. Changes to the execution model must
take these four scenarios into account:

1) Single process. This is the simplest way of executing DAGs. Everything is
executed in the same process (i.e., using the Serial executor with the option
for using subprocesses turned off).

2) Multiple processes, single node. This happens when executing the DAG using
Serial with the option for subprocesses turned on or when using the Parallel
executor. On this case, subprocesses must report back the result of building
a given task to the main process.

3) Multiple nodes, single filesystem. This happens when execution is
orchestrated by another system such as Argo or Airflow. We no longer use
an executor but rather build each task individually. We can no longer monitor
task status as a whole so we rely on the external system for doing so. Since
there is a shared filesystem, upstream products are available to any given
task.

4) Multiple nodes, multiple filesystems. This is the most complex setup.
(e.g., executing in Argo with completely isolated pods). Since File
products from one task aren't available for the next one, the user has to
configure a File.client. When executing each task (Task.build), we use the
client to fetch upstream dependencies and execute a given task.
"""
from functools import reduce
import fnmatch
import sys
import os
from collections.abc import Iterable
import traceback
from copy import copy, deepcopy
from pathlib import Path
import warnings
import logging
import tempfile
from math import ceil
from functools import partial
from ploomber.dag import plot

try:
    import importlib.resources as importlib_resources
except ImportError:  # pragma: no cover
    # backported
    import importlib_resources

import networkx as nx
from tqdm.auto import tqdm
from jinja2 import Template
from IPython.display import Image

from ploomber.table import Table, TaskReport, BuildReport
from ploomber.products import MetaProduct
from ploomber.util import (image_bytes2html, isiterable)
from ploomber.util.debug import debug_if_exception
from ploomber import resources
from ploomber import executors
from ploomber.executors import _format
from ploomber.constants import TaskStatus, DAGStatus
from ploomber.exceptions import (DAGBuildError, DAGRenderError,
                                 DAGBuildEarlyStop, DAGCycle)
from ploomber.messagecollector import (RenderExceptionsCollector,
                                       RenderWarningsCollector)
from ploomber.util.util import callback_check, _make_requires_error_message
from ploomber.dag.dagconfiguration import DAGConfiguration
from ploomber.dag.daglogger import DAGLogger
from ploomber.dag.dagclients import DAGClients
from ploomber.dag.abstractdag import AbstractDAG
from ploomber.dag.util import (check_duplicated_products,
                               fetch_remote_metadata_in_parallel,
                               _path_for_plot)
from ploomber.tasks.abc import Task

if sys.version_info < (3, 8):
    # pygraphviz dropped support for python 3.7
    _conda_cmd = "conda install 'pygraphviz<1.8' -c conda-forge"
else:
    _conda_cmd = "conda install pygraphviz -c conda-forge"

_pygraphviz_message = (
    f"Note that 'pygraphviz' requires 'graphviz' (which is not "
    "pip-installable). To install both: "
    f"{_conda_cmd}\nFor alternatives, see: "
    "https://ploomber.io/s/plot")


class DAG(AbstractDAG):
    """A collection of tasks with dependencies

    Parameters
    ----------
    name : str, optional
        A name to identify this DAG
    clients : dict, optional
        A dictionary with classes as keys and clients as values, can be
        later modified using dag.clients[dag] = client
    differ : CodeDiffer
        An object to determine whether two pieces of code are the same and
        to output a diff, defaults to CodeDiffer() (default parameters)
    executor : str or ploomber.executors instance, optional
        The executor to use (ploomber.executors.Serial and
        ploomber.executors.Parallel), is a string is passed ('serial'
        or 'parallel') the corresponding executor is initialized with default
        parameters

    Attributes
    ----------
    name : str
        A name to identify the DAG

    clients : dict
        A class to client mapping

    executor : ploomber.Executor
        Executor object to run tasks

    on_render : callable
        Function to execute upon rendering. Can request a "dag" parameter.

    on_finish : callable
        Function to execute upon execution. Can request a "dag" parameter
        and/or "report", which contains the report object returned by the
        build function.

    on_failure : callable
        Function to execute upon failure. Can request a "dag" parameter and/or
        "traceback" which will contain a dictionary, possible keys are "build"
        which contains the build error traceback and "on_finish" which contains
        the on_finish hook traceback, if any.

    serializer : callable
        Function to serialize products from PythonCallable tasks. Used if the
        task has no serializer. See ``ploombe.tasks.PythonCallable``
        documentation for details.

    unserializer : callable
        Function to unserialize products from PythonCallable tasks. Used if the
        task has no serializer. See ``ploombe.tasks.PythonCallable``
        documentation for details.
    """
    def __init__(self, name=None, clients=None, executor='serial'):
        self._G = nx.DiGraph()

        self.name = name or 'No name'
        self._logger = logging.getLogger(__name__)

        self._clients = DAGClients(clients)
        self.__exec_status = DAGStatus.WaitingRender

        if executor == 'serial':
            self._executor = executors.Serial()
        elif executor == 'parallel':
            self._executor = executors.Parallel()
        elif isinstance(executor, executors.abc.Executor):
            self._executor = executor
        else:
            raise TypeError(
                'executor must be "serial", "parallel" or '
                'an instance of executors.Executor, got type {}'.format(
                    type(executor)))

        self.on_finish = None
        self.on_failure = None
        self.on_render = None
        self._available_callback_kwargs = {'dag': self}

        self._params = DAGConfiguration()

        # task access differ using .dag.differ
        self.differ = self._params.differ

        self.serializer = None
        self.unserializer = None

    @property
    def executor(self):
        return self._executor

    @executor.setter
    def executor(self, value):
        self._executor = value

    @property
    def _exec_status(self):
        return self.__exec_status

    @_exec_status.setter
    def _exec_status(self, value):
        self._logger.debug('Setting %s status to %s', self, value)

        # The Task class is responsible for updating their status
        # (except for Executed and Errored, those are updated by the executor)
        # DAG should not set task status but only verify that after an attemp
        # to change DAGStatus, all Tasks have allowed states, otherwise there
        # is an error in either the Task or the Executor. we cannot raise an
        # exception here, since setting _exec_status happens might happen
        # right before catching an exception, but we still have to warn the
        # user that the DAG entered an inconsistent state. We only raise
        # an exception when trying to set an invalid value
        # NOTE: in some exec_status, it is ok to raise an exception, maybe we
        # should do it?

        if value == DAGStatus.WaitingRender:
            self.check_tasks_have_allowed_status({TaskStatus.WaitingRender},
                                                 value)

        # render errored
        elif value == DAGStatus.ErroredRender:
            allowed = {
                TaskStatus.WaitingExecution,
                TaskStatus.WaitingUpstream,
                TaskStatus.ErroredRender,
                TaskStatus.AbortedRender,
                TaskStatus.Skipped,
            }
            self.check_tasks_have_allowed_status(allowed, value)

        # rendering ok, waiting execution
        elif value == DAGStatus.WaitingExecution:
            exec_values = set(task.exec_status for task in self.values())
            allowed = {
                TaskStatus.WaitingExecution,
                TaskStatus.WaitingDownload,
                TaskStatus.WaitingUpstream,
                TaskStatus.Skipped,
            }
            self.check_tasks_have_allowed_status(allowed, value)

        # attempted execution but failed
        elif value == DAGStatus.Executed:
            exec_values = set(task.exec_status for task in self.values())
            # check len(self) to prevent this from failing on an empty DAG
            if not exec_values <= {TaskStatus.Executed, TaskStatus.Skipped
                                   } and len(self):
                warnings.warn('The DAG "{}" entered in an inconsistent '
                              'state: trying to set DAG status to '
                              'DAGStatus.Executed but executor '
                              'returned tasks whose status is not '
                              'TaskStatus.Executed nor '
                              'TaskStatus.Skipped, returned '
                              'status: {}'.format(self.name, exec_values))
        elif value == DAGStatus.Errored:
            # no value validation since this state is also set then the
            # DAG executor ends up abrubtly
            pass
        else:
            raise RuntimeError('Unknown DAGStatus value: {}'.format(value))

        self.__exec_status = value

    def check_tasks_have_allowed_status(self, allowed, new_status):
        exec_values = set(task.exec_status for task in self.values())
        if not exec_values <= allowed:
            warnings.warn('The DAG "{}" entered in an inconsistent state: '
                          'trying to set DAG status to '
                          '{} but executor '
                          'returned tasks whose status is not in a '
                          'subet of {}. Returned '
                          'status: {}'.format(self.name, new_status, allowed,
                                              exec_values))

    @property
    def product(self):
        # NOTE: this allows declaring a dag as a dependency for one task,
        # maybe create a metaclass that applies to DAGs and Task
        # We have to rebuild it since tasks might have been added
        return MetaProduct([t.product for t in self.values()])

    @property
    def clients(self):
        return self._clients

    def pop(self, name):
        """Remove a task from the dag
        """
        t = self._G.nodes[name]['task']
        self._G.remove_node(name)
        return t

    def render(self, force=False, show_progress=True, remote=False):
        """
        Render resolves all placeholders in tasks and determines whether
        a task should run or not based on the task.product metadata, this
        allows up-to-date tasks to be skipped

        Parameters
        ----------
        force : bool, default=False
            Ignore product metadata status and prepare all tasks to be
            executed. This option renders much faster in DAGs with products
            whose metadata is stored in remote systems, because there is no
            need to fetch metadata over the network. If the DAG won't be
            built, this option is recommended.

        show_progress : bool, default=True
            Show progress bar

        remote : bool, default=False
            Use remote metadata for determining task status. In most scenarios,
            you want this to be False, Ploomber uses this internally when
            exporting pipelines to other platforms (via Soopervisor).
        """
        g = self._to_graph(fmt='networkx')

        def unique(elements):
            elements_unique = []
            for elem in elements:
                if elem not in elements_unique:
                    elements_unique.append(elem)
            return elements_unique

        dags = unique([t.dag for t in g])

        # first render any other dags involved (this happens when some
        # upstream parameters come form other dags)
        # NOTE: for large compose dags it might be wasteful to render over
        # and over
        for dag in dags:
            if dag is not self:
                dag._render_current(force=force,
                                    show_progress=show_progress,
                                    remote=remote)

        # then, render this dag
        self._render_current(force=force,
                             show_progress=show_progress,
                             remote=remote)

        return self

    def _render_current(self, force, show_progress, remote):
        """
        Render tasks, and update exec_status
        """
        if not self._params.cache_rendered_status:
            # if forcing rendering, there is no need to download metadata
            if not force:
                fetch_remote_metadata_in_parallel(self)

            if show_progress:
                tasks = tqdm(self.values(), total=len(self))
            else:
                tasks = self.values()

            exceptions = RenderExceptionsCollector()
            warnings_ = RenderWarningsCollector()

            # reset all tasks status
            for task in tasks:
                task.exec_status = TaskStatus.WaitingRender

            for t in tasks:
                # no need to process task with AbortedRender
                if t.exec_status == TaskStatus.AbortedRender:
                    continue

                if show_progress:
                    tasks.set_description(
                        'Rendering DAG "{}"'.format(self.name)
                        if self.name != 'No name' else 'Rendering DAG')

                with warnings.catch_warnings(record=True) as warnings_current:
                    warnings.simplefilter("ignore", DeprecationWarning)

                    try:
                        t.render(
                            force=force,
                            outdated_by_code=self._params.outdated_by_code,
                            remote=remote)
                    except Exception as e:
                        tr = _format.exception(e)
                        exceptions.append(task=t, message=tr)

                if warnings_current:
                    warnings_str = [str(w.message) for w in warnings_current]
                    warnings_.append(task=t, message='\n'.join(warnings_str))

            if warnings_:
                # FIXME: maybe raise one by one to keep the warning type
                warnings.warn(str(warnings_))

            if exceptions:
                self._exec_status = DAGStatus.ErroredRender
                raise DAGRenderError(str(exceptions))

            try:
                self._run_on_render()
            except Exception as e:
                # error in hook, log exception
                msg = ('Exception when running on_render '
                       'for DAG "{}": {}'.format(self.name, e))
                self._logger.exception(msg)
                self._exec_status = DAGStatus.ErroredRender
                raise DAGRenderError(msg) from e

            check_duplicated_products(self)

            self._exec_status = DAGStatus.WaitingExecution

    def build(self,
              force=False,
              show_progress=True,
              debug=False,
              close_clients=True):
        """
        Runs the DAG in order so that all upstream dependencies are run for
        every task

        Parameters
        ----------
        force : bool, default=False
            If True, it will run all tasks regardless of status, defaults to
            False

        show_progress : bool, default=True
            Show progress bar

        debug : bool, default=False
            Drop a debugging session if building raises an exception. Note that
            this modifies the executor, temporarily setting it to Serial
            with subprocess off and catching exceptions/warnings off. Restores
            the original executor at the end

        close_clients : bool, default=True
            Close all clients (dag-level, task-level and product-level) upon
            successful build

        Notes
        -----
        All dag-level clients are closed after calling this function

        ``debug`` is useful to let a pipeline run and start debugging at a
        failing  PythonCallable task but it won't work with failing
        ``NotebookRunner`` tasks because notebooks/scripts are executed in a
        different process. If you want to debug ``NotebookRunner`` tasks, use
        ``NotebookRunner.debug()`` instead.

        Returns
        -------
        BuildReport
            A dict-like object with tasks as keys and dicts with task
            status as values
        """
        kwargs = callback_check(self._params.logging_factory,
                                available={'dag_name': self.name})

        res = self._params.logging_factory(**kwargs)

        if isinstance(res, Iterable):
            dag_logger = DAGLogger(*res)
        else:
            dag_logger = DAGLogger(handler=res)

        # if debug, we have to change the executor to these settings, if we run
        # tasks in a subprocess or catch exception, we won't be able to start
        # the debugging session in the right place
        if debug:
            executor_original = self.executor
            self.executor = executors.Serial(build_in_subprocess=False,
                                             catch_exceptions=False,
                                             catch_warnings=False)

        callable_ = partial(self._build,
                            force=force,
                            show_progress=show_progress)

        with dag_logger:
            try:
                if debug:
                    report = debug_if_exception(callable_)
                else:
                    report = callable_()
            finally:
                if close_clients:
                    self.close_clients()

        if debug:
            self.executor = executor_original

        return report

    def _build(self, force, show_progress):
        # always render before building (the function might immediately
        # return if the user turned render status caching on)
        # Do not show progress - should only be displayed when .render is
        # called directly
        self.render(force=force, show_progress=False)

        if self._exec_status == DAGStatus.ErroredRender:
            raise DAGBuildError('Cannot build dag that failed to render, '
                                'fix rendering errors then build again. '
                                'To see the full traceback again, run '
                                'dag.render(force=True)')
        else:
            self._logger.info('Building DAG %s', self)

            tb = {}

            try:
                # within_dag flags when we execute a task in isolation
                # vs as part of a dag execution
                # FIXME: not passing force flag
                task_reports = self._executor(dag=self,
                                              show_progress=show_progress)

            # executors raise this error to signal that there was an error
            # building the dag, this allows us to run the on_failure hook,
            # but any other errors should not be caught (e.g.
            # a user might turn that setting off in the executor to start
            # a debugging session at the line of failure)
            except DAGBuildError as e:
                tb['build'] = traceback.format_exc()
                self._exec_status = DAGStatus.Errored
                build_exception = e
            except DAGBuildEarlyStop:
                # early stop and empty on_failure, nothing left to do
                if self.on_failure is None:
                    return
            else:
                # no error when building dag
                build_exception = None

            if build_exception is None:
                empty = [
                    TaskReport.empty_with_name(t.name) for t in self.values()
                    if t.exec_status == TaskStatus.Skipped
                ]

                build_report = BuildReport(task_reports + empty)
                self._logger.info(' DAG report:\n{}'.format(build_report))

                # try on_finish hook
                try:
                    self._run_on_finish(build_report)
                except Exception as e:
                    tb['on_finish'] = traceback.format_exc()
                    # on_finish error, log exception and set status
                    msg = ('Exception when running on_finish '
                           'for DAG "{}": {}'.format(self.name, e))
                    self._logger.exception(msg)
                    self._exec_status = DAGStatus.Errored

                    if isinstance(e, DAGBuildEarlyStop):
                        # early stop, nothing left to co
                        return
                    else:
                        # otherwise raise exception
                        raise DAGBuildError(msg) from e
                else:
                    # DAG success and on_finish did not raise exception
                    self._exec_status = DAGStatus.Executed
                    return build_report

            else:
                # DAG raised error, run on_failure hook
                try:
                    self._run_on_failure(tb)
                except Exception as e:
                    # error in hook, log exception
                    msg = ('Exception when running on_failure '
                           'for DAG "{}": {}'.format(self.name, e))
                    self._logger.exception(msg)

                    # do not raise exception if early stop
                    if isinstance(e, DAGBuildEarlyStop):
                        return
                    else:
                        raise DAGBuildError(msg) from e

                # on_failure hook executed, raise original exception
                raise build_exception

    def close_clients(self):
        """Close all clients (dag-level, task-level and product-level)
        """
        for client in self.clients.values():
            client.close()

        for task_name in self._iter():
            task = self[task_name]

            if task.client:
                task.client.close()

            if task.product.client:
                task.product.client.close()

    def _run_on_failure(self, tb):
        if self.on_failure:
            self._logger.debug('Executing on_failure hook '
                               'for dag "%s"', self.name)
            kwargs_available = copy(self._available_callback_kwargs)
            kwargs_available['traceback'] = tb

            kwargs = callback_check(self.on_failure, kwargs_available)
            self.on_failure(**kwargs)
        else:
            self._logger.debug('No on_failure hook for dag '
                               '"%s", skipping', self.name)

    def _run_on_finish(self, build_report):
        if self.on_finish:
            self._logger.debug('Executing on_finish hook '
                               'for dag "%s"', self.name)
            kwargs_available = copy(self._available_callback_kwargs)
            kwargs_available['report'] = build_report
            kwargs = callback_check(self.on_finish, kwargs_available)
            self.on_finish(**kwargs)
        else:
            self._logger.debug('No on_finish hook for dag '
                               '"%s", skipping', self.name)

    def _run_on_render(self):
        if self.on_render:
            self._logger.debug('Executing on_render hook '
                               'for dag "%s"', self.name)
            kwargs_available = copy(self._available_callback_kwargs)
            kwargs = callback_check(self.on_render, kwargs_available)
            self.on_render(**kwargs)
        else:
            self._logger.debug('No on_render hook for dag '
                               '"%s", skipping', self.name)

    def _deepcopy_safe(self):
        try:
            return deepcopy(self)
        except Exception as e:
            raise RuntimeError(
                "An error occurred while copying DAG object") from e

    def build_partially(self,
                        target,
                        force=False,
                        show_progress=True,
                        debug=False,
                        skip_upstream=False):
        """Partially build a dag until certain task

        Parameters
        ----------
        target : str
            Name of the target task (last one to build). Can pass a wildcard
            such as 'tasks-*'

        force : bool, default=False
            If True, it will run all tasks regardless of status, defaults to
            False

        show_progress : bool, default=True
            Show progress bar

        debug : bool, default=False
            Drop a debugging session if building raises an exception. Note that
            this modifies the executor and temporarily sets it to Serial
            with subprocess off and catching exceptions/warnings off. Restores
            the original executor at the end.

        skip_upstream : bool, default=False
            If False, includes all upstream dependencies required to build
            target, otherwise it skips them. Note that if this is True and
            it's not possible to build a given task (e.g., missing upstream
            products), this will fail
        """

        # we have to use a deep copy since using a soft one will corrupt
        # task status in subsequent runs

        dag_copy = self._deepcopy_safe()

        # task names are usually str, although this isn't strictly enforced
        if isinstance(target, str) and '*' in target:
            targets = set(fnmatch.filter(self._iter(), target))

            to_include = [
                self[target]._lineage for target in targets
                if self[target]._lineage
            ]

            if to_include:
                lineage = reduce(lambda a, b: a.union(b), to_include)
            else:
                lineage = set()

        else:
            targets = {target}
            lineage = self[target]._lineage or set()

        to_remove = set(dag_copy) - targets - lineage

        for task in to_remove:
            dag_copy.pop(task)

        # clear metadata in the original dag, because building the copy
        # will make it outdated, we have to force reload from disk
        self._clear_metadata()

        if skip_upstream:
            dag_copy.render(force=force, show_progress=show_progress)
            # to prevent call to dag_copy.render() inside dag_copy.build()
            # to overwrite Skipped status
            dag_copy._params.cache_rendered_status = True

            for name, task in dag_copy.items():
                if name not in targets:
                    dag_copy[name].exec_status = TaskStatus.Skipped

        return dag_copy.build(force=force,
                              show_progress=show_progress,
                              debug=debug)

    def status(self, **kwargs):
        """Returns a table with tasks status
        """
        # FIXME: delete this, make dag.render() return this
        self.render()

        return Table([self[name].status(**kwargs) for name in self])

    def to_markup(self, path=None, fmt='html', sections=None):
        """Returns a str (md or html) with the pipeline's description

        Parameters
        ----------
        sections : list
            Which sections to include, possible values are "plot", "status"
            and "source". Defaults to ["plot", "status"]
        """
        sections = sections or ['plot', 'status']

        if fmt not in {'html', 'md'}:
            raise ValueError('fmt must be html or md, got {}'.format(fmt))

        if 'status' in sections:
            status = self.status().to_format('html')
        else:
            status = False

        if 'plot' in sections:
            fd, path_to_plot = tempfile.mkstemp(suffix='.png')
            os.close(fd)
            self.plot(output=path_to_plot)
            plot = image_bytes2html(Path(path_to_plot).read_bytes())
        else:
            plot = False

        template_md = importlib_resources.read_text(resources, 'dag.md')
        out = Template(template_md).render(plot=plot,
                                           status=status,
                                           source='source' in sections,
                                           dag=self)

        if fmt == 'html':
            # importing this requires mistune. we import here since it's
            # an optional dependency
            from ploomber.util import markup
            import mistune

            renderer = markup.HighlightRenderer()
            out = mistune.markdown(out, escape=False, renderer=renderer)

            # add css
            html = importlib_resources.read_text(resources,
                                                 'github-markdown.html')
            out = Template(html).render(content=out)

        if path is not None:
            Path(path).write_text(out)

        return out

    def plot(self, output='embed', include_products=False, backend=None):
        """Plot the DAG

        Parameters
        ----------
        output : str, default='embed'
            Where to save the output (e.g., pipeline.png). If 'embed', it
            returns an IPython image instead.

        include_products : bool, default=False
            If False, each node only contains the task name, if True
            if contains the task name and products. Only available when using
            the pygraphviz backend

        backend : str, default=None
            How to generate the plot, if None it uses pygraphviz if installed,
            otherwise it uses D3 (which doesn't require extra dependencies),
            you can force to use a backend by passing 'pygraphviz' or 'd3'.
        """
        if backend not in {None, 'd3', 'pygraphviz'}:
            raise ValueError("Expected backend to be: None, 'd3' "
                             f"or 'pygraphviz', but got: {backend!r}")

        if plot.choose_backend(backend) == 'd3':
            if include_products:
                raise ValueError("'include_products' is not supported "
                                 "when using the d3 backend. Switch the "
                                 "flag or change to the pypgrahviz backend")

            if output != 'embed':
                suffix = Path(output).suffix

                if suffix != '.html':
                    raise ValueError('Error when using d3 backend: '
                                     'expected a path with '
                                     f'extension .html, but got: {output!r}, '
                                     'please change the extension')

            # FIXME: add tests for this
            self.render()

            G = self._to_graph(fmt='d3', include_products=include_products)

            dag_json = nx.readwrite.json_graph.node_link_data(G)

            with _path_for_plot(path_to_plot=output, fmt='html') as path:
                plot.with_d3(dag_json, output=path)

                if output == 'embed':
                    return plot.embedded_html(path=path)
                else:
                    return path

        elif not plot.check_pygraphviz_installed() and backend == "pygraphviz":
            raise ImportError(
                _make_requires_error_message(
                    ['pygraphviz<1.8'] if sys.version_info <
                    (3, 8) else ['pygraphviz'], 'plot', _pygraphviz_message))

        # use pygraphviz
        with _path_for_plot(path_to_plot=output, fmt='png') as path:
            # attributes docs:
            # https://graphviz.gitlab.io/_pages/doc/info/attrs.html

            G = self._to_graph(fmt='pygraphviz',
                               include_products=include_products)
            G.draw(path, prog='dot', args='-Grankdir=LR')

            if output == 'embed':
                return Image(filename=path)
            else:
                return path

    def _add_task(self, task):
        """Adds a task to the DAG
        """
        if task.name in self._G:
            raise ValueError('DAG already has a task with name '
                             f'{task.name!r}. Pass a "name" '
                             'argument to the task to '
                             'change it')

        if task.name is not None:
            self._G.add_node(task.name, task=task)
        else:
            raise ValueError('Tasks must have a name, got None')

    def _to_graph(self, fmt, only_current_dag=False, include_products=False):
        """
        Converts the DAG to a Networkx DiGraph object. Since upstream
        dependencies are not required to come from the same DAG,
        this object might include tasks that are not included in the current
        object

        Parameters
        ----------
        fmt : 'networkx', 'pygraphviz', or 'd3'
            Output format

        include_products : bool, default=False
            If False, each node only contains the task name, if True
            if contains the task name and products.
        """
        FMT = {'networkx', 'pygraphviz', 'd3'}

        if fmt not in FMT:
            raise ValueError(f'Invalid format, expected one of: {FMT}')

        # https://networkx.github.io/documentation/networkx-1.10/reference/drawing.html
        # http://graphviz.org/doc/info/attrs.html

        # NOTE: delete this, use existing DiGraph object
        G = nx.DiGraph()

        for task in self.values():

            # these formats are used for plotting, so only pass certain task
            # attributes
            if fmt in {'pygraphviz', 'd3'}:
                outdated = task.product._is_outdated()

                # add parameters for graphviz plotting
                color = ('#F08080' if outdated else '#90EE90')

                label = (_task_short_repr(task)
                         if include_products else task.name)

                attr = {
                    'fillcolor': color,
                    'style': 'dashed, filled' if outdated else 'filled',
                    'fontname': 'Helvetica',
                    'fontsize': '16pt',
                    'id': task.name,
                    'label': label
                }
                # graphviz uses the str representation of the node object to
                # distinguish them - by default str(task) returns
                # str(task.product), we have to make sure that when
                # return_graphviz=True, we pass the task id as node, instead
                # of the full task object, otherwise if two products have the
                # same str representation, nodes will clash
                G.add_node(task.name, **attr)

            # when exporting to networkx, we want the task object
            else:
                G.add_node(task)

            def get_task_id(task):
                """
                Determine what to use to identify the edges the task object
                or task name
                """
                return task if fmt == 'networkx' else task.name

            # add edges
            if only_current_dag:
                G.add_edges_from([(get_task_id(up), get_task_id(task))
                                  for up in task.upstream.values()
                                  if up.dag is self])
            else:
                G.add_edges_from([(get_task_id(up), get_task_id(task))
                                  for up in task.upstream.values()])

        if fmt in {'networkx', 'd3'}:
            return G
        else:
            # to_agraph converts to pygraphviz
            return nx.nx_agraph.to_agraph(G)

    def _add_edge(self, task_from, task_to, group_name=None):
        """Add an edge between two tasks

        Parameters
        ----------
        group_name : str
            Pass a string to group this edge, upon rendering, upstream
            products are available via task[group_name][tas_name]
        """
        attrs = {} if group_name is None else {'group_name': group_name}

        # when adding a task group (but not a dag)
        if isiterable(task_from) and not isinstance(task_from, DAG):
            # if iterable, add all components as separate upstream tasks
            for a_task_from in task_from:

                # this happens when the task was originally declared in
                # another dag...
                if a_task_from.name not in self._G:
                    self._G.add_node(a_task_from.name, task=a_task_from)

                self._G.add_edge(a_task_from.name, task_to.name, **attrs)

        else:
            # this happens when the task was originally declared in
            # another dag...
            if task_from.name not in self._G:
                self._G.add_node(task_from.name, task=task_from)

            # DAGs are treated like a single task
            self._G.add_edge(task_from.name, task_to.name, **attrs)

    def _get_upstream(self, task_name):
        """Get upstream tasks given a task name (returns Task objects)
        """
        upstream_names = self._G.predecessors(task_name)
        return {name: self._G.nodes[name]['task'] for name in upstream_names}

    def get_downstream(self, task_name):
        """
        Get downstream tasks for a given task name
        """
        return list(self._G.successors(task_name))

    def _clear_metadata(self):
        """
        Getting product status (outdated/up-to-date) is slow, especially for
        product whose metadata is stored remotely. This is critical when
        rendering because we need to do a forward pass to know which tasks to
        run and a product's status depends on its upstream product's status,
        and we have to make sure we only retrieve metadata once, so we save a
        local copy. But even with this implementation, we don't throw away
        product status after rendering, otherwise calls that need project
        status (like DAG.plot, DAG.status, DAG.to_markup) would have to get
        product status before running its logic, so once we get it, we stick
        with it. The only caveat is that status updates won't be reflected
        immediately (e.g. if the user manually deletes a product's metadata),
        but that's a small price to pay given that this is not expected to
        happen often. The only case when we *must* be sure that we have
        up-to-date metadata is when calling DAG.build(), so we call this
        method before building, which forces metadata reload.
        """
        self._logger.debug('Clearing product status')
        # clearing out this way is only useful after building, but not
        # if the metadata changed since it wont be reloaded
        for task in self.values():
            task.product.metadata.clear()

    def __getitem__(self, key) -> Task:
        try:
            return self._G.nodes[key]['task']
        except KeyError as e:
            e.args = ('DAG does not have a task with name {}'.format(
                repr(key)), )
            raise

    def __delitem__(self, key):
        # TODO: this implementation is correct but perhaps we should raise
        # warning if the deleted task has downstream dependencies, render
        # and build will no longer work
        return self._G.remove_node(key)

    def __iter__(self):
        """
        Iterate task names in topological order. Topological order is
        desirable in many situations, this order guarantees that for any given
        task, its dependencies are executed first, but it's also useful for
        other purposes, such as listing tasks, because it shows a more natural
        order to see how data flows start to finish. For cases where this
        sorting is not required, used the DAG._iter() method instead.


        Notes
        -----
        https://en.wikipedia.org/wiki/Topological_sorting
        """
        # TODO: raise a warning if this any of this dag tasks have tasks
        # from other tasks as dependencies (they won't show up here)
        try:
            for name in nx.algorithms.topological_sort(self._G):
                yield name
        except nx.NetworkXUnfeasible:
            raise DAGCycle

    def _iter(self):
        """Iterate over tasks names (unordered but more efficient than __iter__
        """
        for name in self._G:
            yield name

    def __len__(self):
        return len(self._G)

    def __repr__(self):
        return '{}("{}")'.format(type(self).__name__, self.name)

    # IPython integration
    # https://ipython.readthedocs.io/en/stable/config/integrating.html

    def _ipython_key_completions_(self):
        return list(self)

    # __getstate__ and __setstate__ are needed to make this picklable

    def __getstate__(self):
        state = self.__dict__.copy()
        # _logger is not pickable, so we remove them and build
        # them again in __setstate__
        del state['_logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))


def _single_product_short_repr(product):
    s = repr(product)

    if len(s) > 20:
        s_short = ''

        t = ceil(len(s) / 20)

        for i in range(t):
            s_short += s[(20 * i):(20 * (i + 1))] + '\n'
    else:
        s_short = s

    return s_short


def _meta_product_short_repr(metaproduct):
    return ', '.join(
        [_single_product_short_repr(p) for p in metaproduct.products])


def _product_short_repr(product):
    if isinstance(product, MetaProduct):
        return _meta_product_short_repr(product)
    else:
        return _single_product_short_repr(product)


def _task_short_repr(task):
    def short(s):
        max_l = 30
        return s if len(s) <= max_l else s[:max_l - 3] + '...'

    return ('{} -> \n{}'.format(short(str(task.name)),
                                _product_short_repr(task.product)))



import pickle
try:
    import importlib.resources as importlib_resources
except ImportError:  # pragma: no cover
    # backported
    import importlib_resources
import abc
from pathlib import Path
from itertools import chain

from ploomber import InMemoryDAG, DAG
from ploomber.spec.dagspec import DAGSpecPartial
from ploomber.tasks import input_data_passer, in_memory_callable


class OnlineDAG(abc.ABC):
    """
    Execute partial DAGs in-memory. This is an abstract class, to use it.
    Create a subclass and provide the required static methods.

    See here for a complete example:
    https://github.com/ploomber/projects/blob/master/templates/ml-online/src/ml_online/infer.py
    """

    # FIXME: add a way to customize
    def __init__(self):
        dag = self.init_dag_from_partial(self.get_partial())

        # TODO: add support for manually specifying upstream dependencies
        upstream = {
            name: dag[name].source.extract_upstream()
            for name in dag._iter()
        }

        # names of all tasks used as upstream
        upstream_tasks = chain(*upstream.values())

        # find tasks that are declared as upstream but do not exist in the dag
        missing = set(upstream_tasks) - set(dag)

        for name in missing:
            input_data_passer(dag, name=name)

        # TODO: maybe delete all upstream dependencies and set them again
        # (raise a warning if there are some upstream dependencies?)
        # this doesn't happen when we get a yaml file because we control
        # that using extract_upstream=False but might happen if we receive
        # a DAG object already
        # the dag is complete now, set all upstream dependencies
        for name in dag._iter():
            for dependency in upstream.get(name, []):
                dag[name].set_upstream(dag[dependency])

        # get all terminal nodes and make them a dependency of the  node
        terminal_current = [
            name for name, degree in dag._G.out_degree() if not degree
        ]

        # TODO: extract upstream and make sure they match with the ones in
        # terminal_current
        terminal = in_memory_callable(self.terminal_task,
                                      dag,
                                      name='terminal',
                                      params=self.terminal_params())

        for dependency in terminal_current:
            terminal.set_upstream(dag[dependency])

        self.in_memory = InMemoryDAG(dag)

    @classmethod
    def init_dag_from_partial(cls, partial):
        """Initialize partial returned by get_partial()
        """
        if isinstance(partial, (str, Path)):
            spec = DAGSpecPartial(partial)
            return spec.to_dag()
        elif isinstance(partial, DAG):
            return partial
        else:
            raise TypeError(f'Expected {cls.__name__}.get_partial() to '
                            'return a str, pathlib.Path or ploomber.DAG, '
                            f'got {type(partial).__name__}')

    def predict(self, **kwargs):
        """
        Run the DAG

        Parameters
        ----------
        **kwargs
            One parameter per root task (task with no upstream dependencies)
            in the partial DAG.

        Returns
        -------
        A dictionary with {task_name: returned_value}
        """
        return self.in_memory.build(kwargs)

    @abc.abstractstaticmethod
    def get_partial():
        """
        Must return the location of a partial dag (str or pathlib.Path)
        """
        pass

    @abc.abstractstaticmethod
    def terminal_task(upstream, model):
        """
        Las function to execute. The ``upstream`` parameter contains the
        output of all tasks that have no downstream dependencies
        """
        pass

    @abc.abstractstaticmethod
    def terminal_params():
        """
        Must return a dictionary with parameters passed to ``terminal_task``
        """
        pass


class OnlineModel(OnlineDAG):
    """
    A subclass of :py:mod:`ploomber.OnlineDAG` to provider a simpler interface
    for online DAGs whose terminal task calls ``model.predict``.
    ``OnlineModel`` is initialized with a module following a standard
    structure. Looks for a ``pipeline-features.yaml`` in the module's
    root directory (e.g. ``src/my_module/pipeline-features.yaml``),
    a ``model.pickle`` in the module's root directory. The terminal task
    is executed with a ``model`` parameter which contains the load model
    and calls ``model.predict``. The last task in ``pipeline-features.yaml``
    should be named ``features``.

    See here for a complete example:
    https://github.com/ploomber/projects/blob/master/templates/ml-online/src/ml_online/infer.py

    Parameters
    ----------
    module
        A module following a standard structure

    Examples
    --------
    >>> import my_module # doctest: +SKIP
    >>> model = OnlineModel(my_module) # doctest: +SKIP
    >>> model.predict(x=some_input) # doctest: +SKIP
    """
    def __init__(self, module):
        self._module = module
        super().__init__()

        names = list(self.in_memory.dag._iter())

        if 'features' not in names:
            raise ValueError(
                f'Error initializing {type(self).__name__}. The initialized '
                'DAG must contain a task named "features". '
                f'Got tasks with names: {names}')

    def get_partial(self):
        with importlib_resources.path(
                self._module, 'pipeline-features.yaml') as path_to_spec:
            path = path_to_spec

        return path

    def terminal_params(self):
        error = None

        try:
            content = importlib_resources.read_binary(self._module,
                                                      'model.pickle')
        except FileNotFoundError as e:
            error = e

        if error:
            raise FileNotFoundError(
                f'Error initializing {type(self).__name__} '
                f'from module {self._module.__name__!r}: Missing model file. '
                f'Expected location: {error.filename}')

        model = pickle.loads(content)
        return dict(model=model)

    @staticmethod
    def terminal_task(upstream, model):
        return model.predict(upstream['features'])

    def predict(self, **kwargs):
        """
        Returns the output of ``model.predict(upstream['features'])``
        """
        return self.in_memory.build(kwargs)['terminal']



"""
Implementation of:

$ ploomber cloud

This command runs a bunch of pip/conda commands (depending on what's available)
and it does the *right thing*: creating a new environment if needed, and
locking dependencies.
"""
import os
import json
import uuid
import warnings
from datetime import datetime
from json import JSONDecodeError
import http.client as httplib
import click
from functools import wraps
import re

import humanize

from ploomber.exceptions import BaseException
from ploomber.telemetry import telemetry
from ploomber.telemetry.telemetry import parse_dag, UserSettings

CLOUD_APP_URL = 'api.ploomber.io'
PIPELINES_RESOURCE = '/pipelines'
EMAIL_RESOURCE = '/emailSignup'
headers = {'Content-type': 'application/json'}


def get_key():
    """
    This gets the user cloud api key, returns None if doesn't exist.
    config.yaml is the default user conf file to fetch from.
    """
    if 'PLOOMBER_CLOUD_KEY' in os.environ:
        return os.environ['PLOOMBER_CLOUD_KEY']

    return UserSettings().cloud_key


@telemetry.log_call('set-key')
def set_key(user_key):
    """
    Sets the user cloud api key, if key isn't valid 16 chars length, returns.
    Valid keys are set in the config.yaml (default user conf file).
    """
    _set_key(user_key)


def _set_key(user_key):
    # Validate key
    if not user_key or len(user_key) != 22:
        raise BaseException("The API key is malformed.\n"
                            "Please validate your key or contact the admin.")

    settings = UserSettings()
    settings.cloud_key = user_key
    click.secho("Key was stored")


def get_last_run(timestamp):
    try:
        if timestamp is not None:
            dt = datetime.fromtimestamp(float(timestamp))

            date_h = dt.strftime('%b %d, %Y at %H:%M')
            time_h = humanize.naturaltime(dt)
            last_run = '{} ({})'.format(time_h, date_h)
        else:
            last_run = 'Has not been run'
        return last_run
    except ValueError:
        return timestamp


@telemetry.log_call('get-pipeline')
def get_pipeline(pipeline_id=None, verbose=None):
    """
    Gets a user pipeline via the cloud api key. Validates the key.
    The response is the pipeline instance along with a print statement.
    If the pipeline wasn't found print the server response.
    """
    # Validate API key
    key = get_key()
    if not key:
        return "No cloud API Key was found: {}".format(key)

    # Get pipeline API call
    conn = httplib.HTTPSConnection(CLOUD_APP_URL, timeout=3)
    try:
        headers = {'api_key': key}
        if pipeline_id:
            headers['pipeline_id'] = pipeline_id
        if verbose:
            headers['verbose'] = True
        conn.request("GET", PIPELINES_RESOURCE, headers=headers)

        content = conn.getresponse().read()
        pipeline = json.loads(content)

        for item in pipeline:
            item['updated'] = get_last_run(item['updated'])

        return pipeline
    except JSONDecodeError:
        return "Issue fetching pipeline {}".format(content)
    finally:
        conn.close()


@telemetry.log_call('write-pipeline')
def write_pipeline(pipeline_id,
                   status,
                   log=None,
                   pipeline_name=None,
                   dag=None):
    """
    Updates a user pipeline via the cloud api key. Validates the key.
    The response is the pipeline id if the update was successful.
    If the pipeline wasn't written/updated, the result will contain the error.
    """
    return _write_pipeline(pipeline_id, status, log, pipeline_name, dag)


def _write_pipeline(pipeline_id,
                    status,
                    log=None,
                    pipeline_name=None,
                    dag=None):
    # Validate API key & inputs
    key = get_key()
    if not key:
        return "No cloud API Key was found: {}".format(key)
    if not pipeline_id:
        return "No input pipeline_id: {}".format(key)
    elif not status:
        return "No input pipeline status: {}".format(key)

    # Write pipeline API call
    conn = httplib.HTTPSConnection(CLOUD_APP_URL, timeout=3)
    try:
        headers['api_key'] = key
        body = {
            "pipeline_id": pipeline_id,
            "status": status,
        }
        if pipeline_name:
            body['pipeline_name'] = pipeline_name
        if log:
            body['log'] = log
        if dag:
            body['dag'] = dag
        conn.request("POST",
                     PIPELINES_RESOURCE,
                     body=json.dumps(body),
                     headers=headers)
        res = conn.getresponse()

        content = res.read().decode('utf-8')
        if res.status < 200 or res.status > 300:
            content = f'Issue: {content}'
            warnings.warn(content)

        return content
    except Exception as e:
        return "Issue on fetching pipeline {}".format(e)
    finally:
        conn.close()


@telemetry.log_call('delete-pipeline')
def delete_pipeline(pipeline_id):
    """
    Updates a user pipeline via the cloud api key. Validates the key.
    The response is the pipeline id if the update was successful.
    If the pipeline wasn't written/updated, the result will contain the error.
    """
    # Validate inputs
    key = get_key()
    if not key:
        return "No cloud API Key was found: {}".format(key)
    if not pipeline_id:
        return "No input pipeline_id: {}".format(key)

    # Delete pipeline API call
    conn = httplib.HTTPSConnection(CLOUD_APP_URL, timeout=3)
    try:
        headers['api_key'] = key
        headers['pipeline_id'] = pipeline_id
        conn.request("DELETE", PIPELINES_RESOURCE, headers=headers)

        res = conn.getresponse()
        content = ''
        if res.status < 200 or res.status > 300:
            content += 'Issue: '

        content += res.read().decode('utf-8')
        return content
    except Exception as e:
        return "Issue deleting pipeline {}".format(e)
    finally:
        conn.close()


def cloud_wrapper(payload=False):
    """Runs a function and logs the pipeline status
    """
    def _cloud_call(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _payload = dict()
            pid = str(uuid.uuid4())

            res = str(_write_pipeline(pipeline_id=pid, status='started'))
            if 'Error' in res:
                warnings.warn(res)

            try:
                if payload:
                    result = func(_payload, *args, **kwargs)
                else:
                    result = func(*args, **kwargs)
            except Exception as e:
                res = str(
                    _write_pipeline(pipeline_id=pid,
                                    status='error',
                                    log=str(e.args)))
                if 'Error' in res:
                    warnings.warn(res)
                raise e
            else:

                dag = parse_dag(result)
                res = str(
                    _write_pipeline(pipeline_id=pid,
                                    status='finished',
                                    dag=dag))
                if 'Error' in str(res):
                    warnings.warn(res)
            return result

        return wrapper

    return _cloud_call


def _get_input(text):
    return input(text)


def _email_input():
    # Validate that's the first email registration
    settings = UserSettings()
    if not settings.user_email:
        email = _get_input(
            "\nPlease add your email to get updates and support "
            "(type enter to skip): \n")
        _email_validation(email)


def _email_validation(email):
    pattern = r"[^@]+@[^@]+\.[^@]+"
    settings = UserSettings()
    if re.match(pattern, email):
        # Save in conf file
        settings.user_email = email

        # Call API
        _email_registry(email)
    else:
        # Save in conf file
        settings.user_email = 'empty_email'


def _email_registry(email):
    conn = httplib.HTTPSConnection(CLOUD_APP_URL, timeout=3)
    try:
        user_headers = {'email': email, 'source': 'OS'}
        conn.request("POST", EMAIL_RESOURCE, headers=user_headers)
        print("Thanks for signing up!")
    except httplib.HTTPException:
        pass
    finally:
        conn.close()



from concurrent.futures import ThreadPoolExecutor, as_completed


def download_products_in_parallel(tasks):
    """Call Task.product.download in parallel
    """
    with ThreadPoolExecutor(max_workers=64) as executor:
        future2task = {executor.submit(t.product.download): t for t in tasks}

        for future in as_completed(future2task):
            exception = future.exception()

            if exception:
                task = future2task[future]
                raise RuntimeError(
                    'An error occurred when downloading product from '
                    f'task: {task!r}') from exception



import abc


class Source(abc.ABC):
    """Source abstract class

    Sources encapsulate the code that will be executed. Tasks only focus on
    execution and sources take care of the source code lifecycle: import,
    render (add parameter and validate) then hand off execution to the Task
    object.

    They validation logic is optional and is done at init time to prevent
    Tasks from being instantiated with ill-defined sources. For example, a
    SQLScript is expected to create a table/view, if the source does not
    contain a CREATE TABLE/VIEW statement, then validation could prevent this
    source code from executing. Validation can also happen after rendering.
    For example, a PythonCallableSource checks that the passed Task.params are
    compatible with the source function signature.

    A new Task does not mean a new source is required, the concrete classes
    implemented cover most use cases, if none of the current implementations
    matches their use case, they can either use a GenericSource or implement
    their own.

    Concrete classes should implement its own __repr__ method, which should
    display the source class, code (first few lines if long) and file location,
    if loaded from a file (sources can also load from strings)
    """
    @abc.abstractmethod
    def __init__(self, primitive, hot_reload=False):
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def primitive(self):
        """
        Return the argument passed to build the source, unmodified. e.g. For
        SQL code this is a string, for notebooks it is a JSON string (or
        a plain text code string in a formate that can be converted to a
        notebook), for PythonCallableSource, it is a callable object.

        Should load from disk each time the user calls source.primitive
        if hot_reload is True. Any other object build from the primitive
        (e.g. the code with the injected parameters) should access this
        so hot_reload is propagated.
        """
        # FIXME: there are some API inconsistencies. Most soruces just
        # return the raw argument that initialized them but NotebookSource
        # loads the file if it's a path
        pass  # pragma: no cover

    # TODO: rename to params
    # NOTE: maybe allow dictionaries for cases where default values are
    # possible? python callables and notebooks
    @property
    @abc.abstractmethod
    def variables(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def render(self, params):
        """Render source (fill placeholders)

        If hot_reload is True, this should indirectly reload primitive
        from disk by using self.primitive
        """
        pass  # pragma: no cover

    # NOTE: maybe rename to path? the only case where it is not exactly a path
    # is when it is a callable (line number is added :line), but this is
    # intended as a human-readable property
    @property
    @abc.abstractmethod
    def loc(self):
        """
        Source location. For most cases, this is the path to the file that
        contains the source code. Used only for informative purpose (e.g.
        when doing dag.status())
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def __str__(self):
        """
        Must return the code that will be executed by the Task in a
        human-readable form (even if it's not the actual code sent to the
        task, the only case where this happens currently is for notebooks,
        human-readable would be plain text code but the actual code is an
        ipynb file in JSON), if it is modified by the render step
        (e.g. SQL code with tags), calling this before rendering should
        raise an error.

        This and the task params given an unambiguous definition of which code
        will be run. And it's actually the same code that will be executed
        for all cases except notebooks (where the JSON string is passed to
        the task to execute).
        """
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def doc(self):
        """
        Returns code docstring
        """
        pass  # pragma: no cover

    # TODO: rename to suffix to be consistent with pathlib.Path
    @property
    @abc.abstractmethod
    def extension(self):
        """
        Optional property that should return the file extension, used for
        code normalization (applied before determining wheter two pieces
        or code are different). If None, no normalization is done.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _post_render_validation(self, rendered_value, params):
        """
        Validation function executed after rendering
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _post_init_validation(self, value):
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def name(self):
        """
        Name inferred from the primitive. Used to assign default values to
        Task.name
        """
        pass  # pragma: no cover

    # optional

    def extract_product(self):
        raise NotImplementedError('extract_product is not implemented in '
                                  '{}'.format(type(self).__name__))

    def extract_upstream(self):
        raise NotImplementedError('extract_upstream is not implemented in '
                                  '{}'.format(type(self).__name__))



import platform
import stat
import os
import csv
from datetime import datetime
import json
import shutil
from pathlib import Path
import subprocess
from collections import defaultdict

import click

from ploomber.io.terminalwriter import TerminalWriter
from ploomber.cli.io import command_endpoint
from ploomber.table import Table
from ploomber.telemetry import telemetry
from ploomber.exceptions import BaseException
from ploomber.cli.cloud import _email_input

from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.markup import MarkdownLexer
from pygments import highlight

_URL = 'https://github.com/ploomber/projects'
_DEFAULT_BRANCH = 'master'
_home = Path('~', '.ploomber')

_lexer = MarkdownLexer()
_formatter = TerminalFormatter(bg="dark")


def _find_header(md):
    """Find header markers
    """
    mark = '<!-- end header -->'
    lines = md.splitlines()

    for n, line in enumerate(lines):
        if mark == line:
            return n

    return None


def _skip_header(md):
    line = _find_header(md)

    if line:
        lines = md.splitlines()
        return '\n'.join(lines[line + 1:])
    else:
        return md


def _delete_git_repo(path):
    """
    If on windows, we need to change permissionsto delete the repo
    """
    path_to_repo = Path(path, '.git')
    if os.name == 'nt' and path_to_repo.exists():
        for root, dirs, files in os.walk(path_to_repo):
            for dir_ in dirs:
                os.chmod(Path(root, dir_), stat.S_IRWXU)
            for file_ in files:
                os.chmod(Path(root, file_), stat.S_IRWXU)


def _delete(source, sub):
    return source.replace(sub, '')


def _cleanup_markdown(source):
    source = _delete(source, '<!-- start description -->\n')
    source = _delete(source, '<!-- end description -->\n')
    source = _skip_header(source)
    return source


def _display_markdown(tw, path):
    LINES = 10

    source = _cleanup_markdown(path.read_text())

    lines = source.splitlines()

    top_lines = '\n'.join(lines[:LINES])

    tw.write(highlight(top_lines, _lexer, _formatter))

    if len(lines) > LINES:
        tw.write(f'\n[...{str(path)} continues]\n', yellow=True)


class _ExamplesManager:
    """Class for managing examples data
    """
    def __init__(self, home, branch=None):
        self._home = Path(home).expanduser()
        self._path_to_metadata = self._home / '.metadata'
        self._examples = self._home / 'projects'
        self._branch = branch or _DEFAULT_BRANCH
        self._explicit_branch = branch is not None

    @property
    def home(self):
        return self._home

    @property
    def examples(self):
        return self._examples

    @property
    def path_to_metadata(self):
        return self._path_to_metadata

    @property
    def branch(self):
        return self._branch

    def save_metadata(self, branch):
        timestamp = datetime.now().timestamp()
        metadata = json.dumps(dict(timestamp=timestamp, branch=branch))
        self.path_to_metadata.write_text(metadata)

    def load_metadata(self):
        try:
            return json.loads(self.path_to_metadata.read_text())
        except Exception as e:
            click.echo(f'Error loading metadata: {e}')
            return None

    def clone(self):
        if not self.home.exists():
            self.home.mkdir()

        if self.examples.exists():
            _delete_git_repo(self.examples)
            shutil.rmtree(self.examples)

        try:
            subprocess.run([
                'git',
                'clone',
                '--depth',
                '1',
                '--branch',
                self.branch,
                _URL,
                str(self.examples),
            ],
                           check=True)
        except Exception as e:
            exception = e
        else:
            exception = None

        if exception:
            raise RuntimeError(
                'An error occurred when downloading examples. '
                'Verify git is installed and your internet '
                f'connection. (Error message: {str(exception)!r})')

        self.save_metadata(branch=self.branch)

    def outdated(self):
        metadata = self.load_metadata()

        if metadata:
            timestamp = metadata['timestamp']
            then = datetime.fromtimestamp(timestamp)
            now = datetime.now()
            elapsed = (now - then).days
            is_more_than_one_day_old = elapsed >= 1

            is_different_branch = metadata.get('branch') != self.branch

            if is_more_than_one_day_old:
                click.echo('Examples copy is more than 1 day old...')

            if is_different_branch and self._explicit_branch:
                click.echo('Different branch requested...')

            return is_more_than_one_day_old or (is_different_branch
                                                and self._explicit_branch)
        else:
            click.echo('Cloning...')
            return True

    def path_to(self, name):
        return self.examples / name

    def path_to_readme(self):
        return self.examples / 'README.md'

    def list(self):
        with open(self.examples / '_index.csv',
                  newline='',
                  encoding='utf-8-sig') as f:
            rows = list(csv.DictReader(f))

        categories = json.loads((self.examples / '_category.json').read_text())

        by_category = defaultdict(lambda: [])

        for row in rows:
            category = row.pop('category')
            del row['idx']
            by_category[category].append(row)

        tw = TerminalWriter()

        click.echo(f'Branch: {self.branch}')
        tw.sep('=', 'Ploomber examples', blue=True)
        click.echo()

        for category in sorted(by_category):
            title = category.capitalize()
            description = categories.get(category)

            if description:
                title = f'{title} ({description})'

            tw.sep(' ', title, green=True)
            click.echo()
            click.echo(
                Table.from_dicts(by_category[category]).to_format('simple'))
            click.echo()

        tw.sep('=', blue=True)

        tw.write('\nTo run these examples in free, hosted '
                 f'environment, see instructions at: {_URL}')
        tw.write('\nTo download: ploomber examples -n name -o path\n')
        tw.write('Example: ploomber examples -n templates/ml-basic -o ml\n\n')


@command_endpoint
@telemetry.log_call('examples')
def main(name, force=False, branch=None, output=None):
    """
    Entry point for examples
    """
    manager = _ExamplesManager(home=_home, branch=branch)
    tw = TerminalWriter()

    if not manager.examples.exists() or manager.outdated() or force:
        if not manager.examples.exists():
            click.echo('Local copy does not exist...')
        elif force:
            click.echo('Forcing download...')

        manager.clone()

    if not name:
        manager.list()
    else:
        selected = manager.path_to(name)

        if not selected.exists():
            raise BaseException(
                f'There is no example named {name!r}.\n'
                'List examples: ploomber examples\n'
                'Update local copy: ploomber examples -f\n'
                'Get ML example: ploomber examples -n '
                'templates/ml-basic -o ml-example',
                type_='no-example-with-name')
        else:
            output = output or name

            tw.sep('=', f'Copying example {name} to {output}/', blue=True)

            if Path(output).exists():
                raise BaseException(
                    f"{output!r} already exists in the current working "
                    "directory, please rename it or move it "
                    "to another location and try again.",
                    type_='directory-exists')

            shutil.copytree(selected, output)

            path_to_readme = Path(output, 'README.md')
            out_dir = output + ('\\'
                                if platform.system() == 'Windows' else '/')

            tw.write('Next steps:\n\n'
                     f'$ cd {out_dir}'
                     f'\n$ ploomber install')
            tw.write(f'\n\nOpen {str(path_to_readme)} for details.\n',
                     blue=True)
    _email_input()



"""
One of Ploomber's main goals is to allow writing robust/reliable code in an
interactive way. Interactive workflows make people more productive but they
might come in detriment of writing high quality code (e.g. developing a
pipeline in a single ipynb file). The basic idea for this module is to provide
a way to transparently go back and forth between a Task in a DAG and a
temporary Jupyter notebook. Currently, we only provide this for PythonCallable
and NotebookRunner but the idea is to expand to other tasks, so we have to
decide on a common behavior for this, here are a few rules:

1) Temporary jupyter notebook are usually destroyed when the user closes the
jupyter applciation. But there are extraordinary cases where we don't want to
remove it, as it might cause code loss. e.g. if the user calls
PythonCallable.develop() and while it is editing the notebook the module where
the source function is defined, we risk corrupting the module file, so we abort
overriding changes but still keep the temporary notebook. For this reason,
we save temporary notebooks in the same location of the source being edited,
to make it easier to recognize which file is related to.

2) The current working directory (cwd) in the session where Task.develop() is
called can be different from the cwd in the Jupyter application. This happens
because Jupyter sets the cwd to the current parent folder, this means that
any relative path defined in the DAG, will break if the cwd in the Jupyter app
is not the same as in the DAg declaration. To fix this, we always add a top
cell in temporary notebooks to make the cwd the same folder where
Task.develop() was called.

3) [TODO] all temporary cells must have a tmp- preffx


TODO: move the logic that implements NotebookRunner.{develop, debug} to this
module
"""
import importlib
from itertools import chain
from pathlib import Path
import inspect
import warnings

import jupyter_client
# papermill is importing a deprecated module from pyarrow
with warnings.catch_warnings():
    warnings.simplefilter('ignore', FutureWarning)
    from papermill.translators import PythonTranslator
import parso
import nbformat

from ploomber.util import chdir_code
from ploomber.sources.nb_utils import find_cell_with_tag
from ploomber.static_analysis.python import PythonCallableExtractor
from ploomber.sources.inspect import getfile

# TODO: test for locally defined objects
# TODO: reloading the fn causes trobule if it enters into an inconsistent
# state, e.g. a module that does not exist is saved, next time is realoaded,
# it will fail because it has to import such module
# TODO: if we remove upstream refernces from the functions body from jupyter
# the parameter is deleted from the signature but on reload (dag.render())
# signature validation fails bc it still loads the old signature, two options:
# either force reload all modules from all pythoncallables, or re-implement
# the signature check to get the signature using static analysis, not sure
# which is best


class CallableInteractiveDeveloper:
    """Convert callables to notebooks, edit and save back

    Parameters
    ----------
    fn : callable
        Function to edit
    params : dict
        Parameters to call the function

    Examples
    --------
    >>> with CallableInteractiveDeveloper(fn, # doctest: +SKIP
    ...     {'param': 1}) as path_to_nb:
    ...     # do stuff with the notebook file
    ...     pass
    """
    def __init__(self, fn, params):
        self.fn = fn
        self.path_to_source = Path(inspect.getsourcefile(fn))
        self.params = params
        self.tmp_path = self.path_to_source.with_name(
            self.path_to_source.with_suffix('').name + '-tmp.ipynb')
        self._source_code = None

    def _reload_fn(self):
        # force to reload module to get the right information in case the
        # original source code was modified and the function is no longer in
        # the same position
        # NOTE: are there any  problems with this approach?
        # we could also read the file directly and use ast/parso to get the
        # function's information we need
        mod = importlib.reload(inspect.getmodule(self.fn))
        self.fn = getattr(mod, self.fn.__name__)

    def to_nb(self, path=None):
        """
        Converts the function to is notebook representation, Returns a
        notebook object, if path is passed, it saves the notebook as well
        Returns the function's body in a notebook (tmp location), inserts
        params as variables at the top
        """
        self._reload_fn()

        body_elements, _ = parse_function(self.fn)
        top, local, bottom = extract_imports(self.fn)
        return function_to_nb(body_elements, top, local, bottom, self.params,
                              self.fn, path)

    def overwrite(self, obj):
        """
        Overwrite the function's body with the notebook contents, excluding
        injected parameters and cells whose first line is "#". obj can be
        either a notebook object or a path
        """
        self._reload_fn()

        if isinstance(obj, (str, Path)):
            nb = nbformat.read(obj, as_version=nbformat.NO_CONVERT)
        else:
            nb = obj

        nb.cells = nb.cells[:last_non_empty_cell(nb.cells)]

        # remove cells that are only needed for the nb but not for the function
        code_cells = [c['source'] for c in nb.cells if keep_cell(c)]

        # add 4 spaces to each code cell, exclude white space lines
        code_cells = [indent_cell(code) for code in code_cells]

        # get the original file where the function is defined
        content = self.path_to_source.read_text()
        content_lines = content.splitlines()
        trailing_newline = content[-1] == '\n'

        # an upstream parameter
        fn_starts, fn_ends = function_lines(self.fn)

        # keep the file the same until you reach the function definition plus
        # an offset to account for the signature (which might span >1 line)
        _, body_start = parse_function(self.fn)
        keep_until = fn_starts + body_start
        header = content_lines[:keep_until]

        # the footer is everything below the end of the original definition
        footer = content_lines[fn_ends:]

        # if there is anything at the end, we have to add an empty line to
        # properly end the function definition, if this is the last definition
        # in the file, we don't have to add this
        if footer:
            footer = [''] + footer

        new_content = '\n'.join(header + code_cells + footer)

        # replace old top imports with new ones
        new_content_lines = new_content.splitlines()
        _, line = extract_imports_top(parso.parse(new_content),
                                      new_content_lines)
        imports_top_cell, _ = find_cell_with_tag(nb, 'imports-top')

        # ignore trailing whitespace in top imports cell but keep original
        # amount of whitespace separating the last import and the first name
        # definition
        content_to_write = (imports_top_cell['source'].rstrip() + '\n' +
                            '\n'.join(new_content_lines[line - 1:]))

        # if the original file had a trailing newline, keep it
        if trailing_newline:
            content_to_write += '\n'

        # NOTE: this last part parses the code several times, we can improve
        # performance by only parsing once
        m = parso.parse(content_to_write)
        fn_def = find_function_with_name(m, self.fn.__name__)
        fn_code = fn_def.get_code()

        has_upstream_dependencies = PythonCallableExtractor(
            fn_code).extract_upstream()
        upstream_in_func_sig = upstream_in_func_signature(fn_code)

        if not upstream_in_func_sig and has_upstream_dependencies:
            fn_code_new = add_upstream_to_func_signature(fn_code)
            content_to_write = _replace_fn_source(content_to_write, fn_def,
                                                  fn_code_new)

        elif upstream_in_func_sig and not has_upstream_dependencies:
            fn_code_new = remove_upstream_to_func_signature(fn_code)
            content_to_write = _replace_fn_source(content_to_write, fn_def,
                                                  fn_code_new)

        self.path_to_source.write_text(content_to_write)

    def __enter__(self):
        self._source_code = self.path_to_source.read_text()
        self.to_nb(path=self.tmp_path)
        return str(self.tmp_path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        current_source_code = self.path_to_source.read_text()

        if self._source_code != current_source_code:
            raise ValueError(f'File "{self.path_to_source}" (where '
                             f'callable "{self.fn.__name__}" is defined) '
                             'changed while editing the function in the '
                             'notebook app. This might lead to corrupted '
                             'source files. Changes from the notebook were '
                             'not saved back to the module. Notebook '
                             f'available at "{self.tmp_path}')

        self.overwrite(self.tmp_path)
        Path(self.tmp_path).unlink()

    def __del__(self):
        tmp = Path(self.tmp_path)
        if tmp.exists():
            tmp.unlink()


def last_non_empty_cell(cells):
    """Returns the index + 1 for the last non-empty cell
    """
    idx = len(cells)

    for cell in cells[::-1]:
        if cell.source:
            return idx

        idx -= 1

    return idx


def keep_cell(cell):
    """
    Rule to decide whether to keep a cell or not. This is executed before
    converting the notebook back to a function
    """
    cell_tags = set(cell['metadata'].get('tags', {}))

    # remove cell with this tag, they are not part of the function body
    tags_to_remove = {
        'injected-parameters',
        'imports-top',
        'imports-local',
        'imports-bottom',
        'debugging-settings',
    }

    has_tags_to_remove = len(cell_tags & tags_to_remove)

    return (cell['cell_type'] == 'code' and not has_tags_to_remove
            and cell['source'][:2] != '#\n')


def indent_line(lline):
    return '    ' + lline if lline else ''


def indent_cell(code):
    return '\n'.join([indent_line(line) for line in code.splitlines()])


def body_elements_from_source(source):
    # getsource adds a new line at the end of the the function, we don't need
    # this

    body = parso.parse(source).children[0].children[-1]

    # parso is adding a new line as first element, not sure if this
    # happens always though
    if isinstance(body.children[0], parso.python.tree.Newline):
        body_elements = body.children[1:]
    else:
        body_elements = body.children

    return body_elements, body.start_pos[0] - 1


def parse_function(fn):
    """
    Extract function's source code, parse it and return function body
    elements along with the # of the last line for the signature (which
    marks the beginning of the function's body) and all the imports
    """
    # TODO: exclude return at the end, what if we find more than one?
    # maybe do not support functions with return statements for now
    source = inspect.getsource(fn).rstrip()
    body_elements, start_pos = body_elements_from_source(source)
    return body_elements, start_pos


def extract_imports(fn):
    source = Path(getfile(fn)).read_text()

    module = parso.parse(source)
    lines = source.splitlines()

    imports_top, line = extract_imports_top(module, lines)

    # any imports below the top imports
    lines_bottom = '\n'.join(lines[line - 1:])
    imports_bottom = '\n'.join(
        imp.get_code() for imp in parso.parse(lines_bottom).iter_imports())

    # generate imports from local definitions
    imports_local = make_import_from_definitions(module, fn)

    return (
        imports_top,
        imports_local,
        imports_bottom if imports_bottom else None,
    )


def extract_imports_top(module, lines):
    ch = module.children[0]

    while True:
        if ch:
            if not has_import(ch):
                break
        else:
            break

        ch = ch.get_next_sibling()

    line, _ = ch.start_pos

    # line numbers start at 1...
    imports_top = '\n'.join(lines[:line - 1])
    new_lines = trailing_newlines(imports_top)

    return imports_top[:-new_lines], line - new_lines


def has_import(stmt):
    """
    Check if statement contains an import
    """
    for ch in stmt.children:
        if ch.type in {'import_name', 'import_from'}:
            return True
    return False


def trailing_newlines(s):
    n = 0

    for char in reversed(s):
        if char != '\n':
            break
        n += 1

    return n


def function_lines(fn):
    lines, start = inspect.getsourcelines(fn)
    end = start + len(lines)
    return start, end


def get_func_and_class_names(module):
    return [
        defs.name.get_code().strip()
        for defs in chain(module.iter_funcdefs(), module.iter_classdefs())
    ]


def make_import_from_definitions(module, fn):
    module_name = inspect.getmodule(fn).__name__
    names = [
        name for name in get_func_and_class_names(module)
        if name != fn.__name__
    ]

    if names:
        names_all = ', '.join(names)
        return f'from {module_name} import {names_all}'


def function_to_nb(body_elements, imports_top, imports_local, imports_bottom,
                   params, fn, path):
    """
    Save function body elements to a notebook
    """
    # TODO: Params should implement an option to call to_json_serializable
    # on product to avoid repetition I'm using this same code in notebook
    # runner. Also raise error if any of the params is not
    # json serializable
    try:
        params = params.to_json_serializable()
        params['product'] = params['product'].to_json_serializable()
    except AttributeError:
        pass

    nb_format = nbformat.versions[nbformat.current_nbformat]
    nb = nb_format.new_notebook()

    # get the module where the function is declared
    tokens = inspect.getmodule(fn).__name__.split('.')
    module_name = '.'.join(tokens[:-1])

    # add cell that chdirs for the current working directory
    # add __package__, we need this for relative imports to work
    # see: https://www.python.org/dev/peps/pep-0366/ for details
    source = """
# Debugging settings (this cell will be removed before saving)
# change the current working directory to the one when .debug() happen
# to make relative paths work
import os
{}
__package__ = "{}"
""".format(chdir_code(Path('.').resolve()), module_name)
    cell = nb_format.new_code_cell(source,
                                   metadata={'tags': ['debugging-settings']})
    nb.cells.append(cell)

    # then add params passed to the function
    cell = nb_format.new_code_cell(PythonTranslator.codify(params),
                                   metadata={'tags': ['injected-parameters']})
    nb.cells.append(cell)

    # first three cells: imports
    for code, tag in ((imports_top, 'imports-top'),
                      (imports_local, 'imports-local'), (imports_bottom,
                                                         'imports-bottom')):
        if code:
            nb.cells.append(
                nb_format.new_code_cell(source=code,
                                        metadata=dict(tags=[tag])))

    for statement in body_elements:
        lines, newlines = split_statement(statement)

        # find indentation # of characters using the first line
        idx = indentation_idx(lines[0])

        # remove indentation from all function body lines
        lines = [line[idx:] for line in lines]

        # add one empty cell per leading new line
        nb.cells.extend(
            [nb_format.new_code_cell(source='') for _ in range(newlines)])

        # add actual code as a single string
        cell = nb_format.new_code_cell(source='\n'.join(lines))
        nb.cells.append(cell)

    k = jupyter_client.kernelspec.get_kernel_spec('python3')

    nb.metadata.kernelspec = {
        "display_name": k.display_name,
        "language": k.language,
        "name": 'python3'
    }

    if path:
        nbformat.write(nb, path)

    return nb


def split_statement(statement):
    code = statement.get_code()

    newlines = 0

    for char in code:
        if char != '\n':
            break

        newlines += 1

    lines = code.strip('\n').split('\n')
    return lines, newlines


def indentation_idx(line):
    idx = len(line) - len(line.lstrip())
    return idx


def upstream_in_func_signature(source):
    _, params = _get_func_def_and_params(source)
    return 'upstream' in set(p.name.get_code().strip() for p in params
                             if p.type == 'param')


def add_upstream_to_func_signature(source):
    fn, params = _get_func_def_and_params(source)
    # add a "," if there is at least one param
    params.insert(-1, ', upstream' if len(params) > 2 else 'upstream')
    signature = try_get_code(params)
    fn.children[2] = signature
    # delete leading newline code, to avoid duplicating it
    return try_get_code(fn.children).lstrip('\n')


def remove_upstream_to_func_signature(source):
    fn, params = _get_func_def_and_params(source)
    params_names = (p.get_code().strip(', ') for p in params[1:-1])
    params_list = ', '.join(p for p in params_names if p != 'upstream')
    signature = f'({params_list})'
    fn.children[2] = signature
    # delete leading newline code, to avoid duplicating it
    return try_get_code(fn.children).lstrip('\n')


def _get_func_def_and_params(source):
    fn = parso.parse(source).children[0]

    if fn.type != 'funcdef':
        raise ValueError('Expected first element from parse source'
                         f' code to be "funcdef", got {fn.type!r}')

    return fn, fn.children[2].children


def _replace_fn_source(content_to_write, fn_def, fn_code_new):
    line_from, line_to = fn_def.start_pos[0], fn_def.end_pos[0]
    lines = content_to_write.splitlines()
    lines_new = (lines[:line_from - 1] + [fn_code_new] + lines[line_to - 1:])
    return '\n'.join(lines_new)


def try_get_code(elements):
    code = []

    for p in elements:
        try:
            s = p.get_code()
        except AttributeError:
            s = p

        code.append(s)

    return ''.join(code)


def find_function_with_name(module, fn_name):
    for fn_def in module.iter_funcdefs():
        if fn_def.name.get_code().strip() == fn_name:
            return fn_def



"""
Internal module for making requests to external APIs. It wraps the requests
module to provide friendlier error messages and defaults
"""
import requests

from ploomber.exceptions import NetworkException


def _make_error(reason):
    return NetworkException(f'{reason}\n'
                            'If you need help, reach out to us: '
                            'https://ploomber.io/community')


def _request(method, url, params=None, **kwargs):
    try:
        response = method(url, params=params, **kwargs)
    except requests.exceptions.Timeout:
        # we do this instead of a chained exception since we want to hide
        # the original error and replace it with this one
        error = _make_error('The server took too long to respond '
                            f'when calling: {url}. Try again. ')
    except requests.exceptions.ConnectionError:
        error = _make_error('A connection '
                            f'error occurred when calling: {url}. '
                            'Verify your internet connection. ')
    else:
        error = None

    if error:
        raise error

    if response.status_code >= 300:
        json_ = response.json()
        message = json_.get("Message")

        if message:
            raise NetworkException(
                f'{message} (status: {response.status_code})')
        else:
            raise NetworkException(f'Error: {json_}')

    return response


def get(*args, **kwargs):
    return _request(requests.get, *args, **kwargs)


def post(*args, **kwargs):
    return _request(requests.post, *args, **kwargs)


def put(*args, **kwargs):
    return _request(requests.put, *args, **kwargs)


def delete(*args, **kwargs):
    return _request(requests.delete, *args, **kwargs)



from enum import Enum


# TODO: rename to use the names as the hooks: finished and failure
class DAGStatus(Enum):
    WaitingRender = 'waiting_render'
    WaitingExecution = 'waiting_execution'
    ErroredRender = 'errored_render'
    Executed = 'executed'
    Errored = 'errored'


# TODO: rename to use the names as the hooks: finished and failure
class TaskStatus(Enum):
    # INIT

    # waiting render
    WaitingRender = 'waiting_render'

    # SET AT RENDERING TIME
    # At rendering time: traverse in topological order and determine
    # status by checking downstream dependencies at the end of Task.render

    # up-to-date, no need to run. Propagate to move downstream dependencies
    # from WaitingUpstream to WaitingExecution
    # TODO: rename to UpToDate
    Skipped = 'skipped'

    # outdated: waiting to be run by the executor
    WaitingExecution = 'waiting_execution'
    # outdated: waits for upstream dependencies to be either Executed
    # or Skipped
    WaitingUpstream = 'waiting_upstream'
    # up-to-date but need to download from remote location
    WaitingDownload = 'waiting_download'

    # failed to render, propagate ErroredRender to downstream dependencies
    # downstream dependencies and set them to AbortedRender
    ErroredRender = 'errored_render'
    # Am I really stopping render due to this?
    AbortedRender = 'aborted_render'

    # SET AT BUILDING TIME
    # Two options, one is similar to the process at rendering time: determine
    # status when entering Task.build and by checking downstream dependencies
    # Second option: propagate on task.exec_status = TaskStatus.Status
    # this is less efficient but has the advantage that we could do "live
    # monitoring" as changes in status are always updated

    # succesfully executed, propagate to move downstream dependencies from
    # WaitingUpstream to WaitingExecution
    Executed = 'executed'

    # crashed, propagate to move downstream dependencies from WaitingUpstream
    # to Aborted
    Errored = 'errored'

    # an upstream dependency crashed, do not execute, propagate to abort
    # any downstream dependency
    Aborted = 'aborted'

    BrokenProcessPool = 'broken_process_pool'



from ploomber.cloud import api


def download_data(key):
    """

    Examples
    --------
    >>> from ploomber.cloud import download_data
    >>> download_data('raw.csv') # doctest: +SKIP
    """
    # TODO: this should find the root directory and download relative to it
    # then return the absolute path
    url = api.download_data(key).text
    return api._download_file(url, skip_if_exists=True, raise_on_missing=True)



import typing as t
from click.exceptions import ClickException
from click._compat import get_text_stderr
from click.utils import echo
from gettext import gettext as _


def _format_message(exception):
    if hasattr(exception, 'format_message'):
        return exception.format_message()
    else:
        return str(exception)


def _build_message(exception):
    msg = _format_message(exception)

    while exception.__cause__:
        msg += f'\n{_format_message(exception.__cause__)}'
        exception = exception.__cause__

    return msg


class BaseException(ClickException):
    """
    A subclass of ClickException that adds support for printing error messages
    from chained exceptions
    """
    def __init__(self, message, type_=None):
        super().__init__(message)
        self.type_ = type_

    def get_message(self):
        return f'Error: {_build_message(self)}'

    def show(self, file: t.Optional[t.IO] = None) -> None:
        if file is None:
            file = get_text_stderr()

        echo(_(self.get_message()), file=file)


class DAGRenderError(Exception):
    """Raise when a dag fails to render

    Notes
    -----
    This is a special exception that should only be raised under specific
    circumstances in the DAG implementation. Review carefully since this
    exceptions signals special output formatting in the CLI (via the
    @cli_endpoint decorator)
    """
    def __init__(self, message):
        message = message + '\nNeed help? https://ploomber.io/community'
        super().__init__(message)


class DAGBuildError(Exception):
    """Raise when a dag fails to build

    Notes
    -----
    This is a special exception that should only be raised under specific
    circumstances in the DAG implementation. Review carefully since this
    exceptions signals special output formatting in the CLI (via the
    @cli_endpoint decorator)
    """
    def __init__(self, message):
        message = message + '\nNeed help? https://ploomber.io/community'
        super().__init__(message)


class DAGWithDuplicatedProducts(BaseException):
    """Raised when more than one task has the same product
    """
    pass


class DAGBuildEarlyStop(Exception):
    """
    This is raised on purpose to signal that the DAG should not continue
    executing but is not considered a build error
    """
    pass


class TaskInitializationError(BaseException):
    """Raised when a task fails to initialize
    """
    pass


class TaskBuildError(Exception):
    """Raise when a task fails to build
    """
    pass


class TaskRenderError(Exception):
    """Raise when a task fails to render
    """
    pass


class RenderError(Exception):
    """Raise when a template fails to render
    """
    pass


class SourceInitializationError(BaseException):
    """Raise when a source fails to initialize due to wrong parameters
    """
    pass


class MissingParametersCellError(SourceInitializationError):
    """Raise when a script or notebook is missing the parameters cell
    """
    pass


class CallbackSignatureError(Exception):
    """When a callback function does not have the right signature
    """
    pass


class CallbackCheckAborted(Exception):
    """
    Used by callback_check to signal that signature check is unfeasible because
    the user passed a DottedPath whose underlying function hasn't been imported
    """
    pass


class UpstreamKeyError(Exception):
    """
    Raised when trying to get an upstream dependency that does not exist,
    we have to implement a custom exception otherwise jinja is going
    to ignore our error messages (if we raise the usual KeyError).
    See: https://jinja.palletsprojects.com/en/2.11.x/templates/#variables
    """
    pass


class DAGSpecInitializationError(BaseException):
    """
    Raised when failing to initialize a DAGSpec object
    """
    pass


class DAGSpecInvalidError(Exception):
    """
    Raised when trying to find dagspec automatically but the file doesn't exist
    or there is an invalid configuration
    """
    pass


class DAGCycle(Exception):
    """
    Raised when a DAG is defined with cycles.
    """
    def __init__(self):
        error_message = """
        Failed to process DAG because it contains cycles.
        """
        super().__init__(error_message)


class SpecValidationError(Exception):
    """
    Raised when failing to validate a spec
    """
    def __init__(self, errors, model, kwargs):
        self.errors = errors
        self.model = model
        self.kwargs = kwargs

    def __str__(self):
        n_errors = len(self.errors)

        msg = (f'{n_errors} error{"" if n_errors == 1 else "s"} found '
               f'when validating {self.model.__name__} with values '
               f'{self.kwargs}\n\n'
               f'{display_errors(self.errors)}')

        return msg


class SQLTaskBuildError(TaskBuildError):
    """
    Raised by SQLScript and SQLDump when the .execute method fails
    """
    def __init__(self, type_, source_code, original):
        self.type_ = type_
        self.source_code = source_code
        self.original = original
        error_message = ('An error occurred when executing '
                         f'{type_.__name__} task with '
                         f'source code:\n\n{source_code!r}\n')
        super().__init__(error_message)


class RemoteFileNotFound(Exception):
    """
    Raised by File clients when atempting to download a file that doesn't exist
    """
    pass


class MissingClientError(Exception):
    """
    Raised when failing to get a valid task-level or dag-level client
    for a Task or Product
    """
    pass


class ValidationError(BaseException):
    """Raised when failed to validate input data
    """
    pass


class NetworkException(BaseException):
    """Raised when failin to call remote APIs
    """
    pass


def display_errors(errors):
    return '\n'.join(f'{_display_error_loc(e)} ({e["msg"]})' for e in errors)


def _display_error_loc(error):
    return ' -> '.join(str(e) for e in error['loc'])



from difflib import get_close_matches

from ploomber import tasks
from ploomber import products

_NORMALIZED_TASKS = {name.upper(): name for name in tasks.__all__}
_NORMALIZED_PRODUCTS = {name.upper(): name for name in products.__all__}
_NORMALIZED = {**_NORMALIZED_TASKS, **_NORMALIZED_PRODUCTS}

_KEY2CLASS_TASKS = {name: getattr(tasks, name) for name in tasks.__all__}
_KEY2CLASS_PRODUCTS = {
    name: getattr(products, name)
    for name in products.__all__
}
_KEY2CLASS = {**_KEY2CLASS_TASKS, **_KEY2CLASS_PRODUCTS}


def _suggest_class_name(name: str, options):
    name = name.upper()

    close_commands = get_close_matches(name, options)

    if close_commands:
        return options[close_commands[0]]
    else:
        return None


def _normalize_input(name):
    return name.upper().replace('-', '').replace('_', '').replace(' ', '')


def get_suggestion(key_str, mapping=None):
    key_str_normalized = _normalize_input(key_str)
    mapping = mapping or _NORMALIZED
    return _suggest_class_name(key_str_normalized, mapping)


def str_to_class(key_str):
    return _KEY2CLASS.get(key_str, None)


def validate_task_class_name(value):
    """
    Validates if a string is a valid Task class name (e.g., SQLScipt).
    Raises a ValueError if not.
    """
    if value not in _KEY2CLASS_TASKS:
        suggestion = get_suggestion(value, mapping=_NORMALIZED_TASKS)
        msg = f'{value!r} is not a valid Task class name'

        if suggestion:
            msg += f'. Did you mean {suggestion!r}?'

        raise ValueError(msg)

    return _KEY2CLASS_TASKS[value]


def validate_product_class_name(value):
    """
    Validates if a string is a valid Product class name (e.g., File).
    Raises a ValueError if not.
    """
    if value not in _KEY2CLASS_PRODUCTS:
        suggestion = get_suggestion(value, mapping=_NORMALIZED_PRODUCTS)
        msg = f'{value!r} is not a valid Product class name'

        if suggestion:
            msg += f'. Did you mean {suggestion!r}?'

        raise ValueError(msg)

    return _KEY2CLASS_PRODUCTS[value]



import networkx as nx
from itertools import chain


class SuperDAG:
    """DAG-like interface to make multiple dags behave like one
    """
    def __init__(self, dags):
        G = nx.DiGraph()

        for task in chain(*(dag.values() for dag in dags)):
            G.add_node(task)
            for upstream in task.upstream:
                G.add_edge(task.dag[upstream], task)

        dags = []

        # this only works for simple cases: where sink nodes from one dag
        # are upstream dependencies for root nodes from the next one,
        # it's enough for now - another would be to call build on each
        # task individually but that's gonna make this implementation much
        # more complicated
        for task in nx.algorithms.topological_sort(G):
            if task.dag not in dags:
                dags.append(task.dag)

        self.G = G
        self.dags = dags

    def plot(self, path):
        G_ = nx.nx_agraph.to_agraph(self.G)
        G_.draw(path, prog='dot', args='-Grankdir=LR')

    def build(self):
        for dag in self.dags:
            dag.build()

    def render(self):
        for dag in self.dags:
            dag.render()



from pathlib import Path
from io import StringIO

from jinja2 import Template

from ploomber.tasks.abc import Task
from ploomber.tasks.mixins import ClientMixin
from ploomber.sources import (SQLScriptSource, SQLQuerySource, FileSource)
from ploomber.products import (File, PostgresRelation, SQLiteRelation,
                               GenericSQLRelation, GenericProduct, SQLRelation)
from ploomber import io
from ploomber.util import requires
from ploomber.placeholders.placeholder import _add_globals
from ploomber.exceptions import SQLTaskBuildError


class SQLScript(ClientMixin, Task):
    """Execute a script in a SQL database to create a relation or view

    Parameters
    ----------
    source: str or pathlib.Path
        SQL script source, if str, the content is interpreted as the actual
        script, if pathlib.Path, the content of the file is loaded
    product: ploomber.products.product
        Product generated upon successful execution
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    client: ploomber.clients.{SQLAlchemyClient, DBAPIClient}, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]
    params: dict, optional
        Parameters to pass to the script, by default, the callable will
        be executed with a "product" (which will contain the product object).
        It will also include a "upstream" parameter if the task has upstream
        dependencies along with any parameters declared here. The source
        code is converted to a jinja2.Template for passing parameters,
        refer to jinja2 documentation for details
    """
    PRODUCT_CLASSES_ALLOWED = (PostgresRelation, SQLiteRelation,
                               GenericSQLRelation, SQLRelation)

    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 client=None,
                 params=None):
        params = params or {}
        # TODO: access self.client so it uses the dag-level if available
        try:
            split_source = client.split_source
        except AttributeError:
            split_source = None

        kwargs = dict(hot_reload=dag._params.hot_reload,
                      split_source=split_source)

        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)
        self._client = client
        self.dag = dag

    def run(self):
        source_code = str(self.source)

        try:
            return self.client.execute(source_code)
        except Exception as e:
            raise SQLTaskBuildError(type(self), source_code, e) from e

    def load(self, limit=10):
        """Load this task's product in a pandas.DataFrame

        Parameters
        ----------
        limit : int, default=10
            How many records to load, defaults to 10
        """
        import pandas as pd
        return pd.read_sql(f'SELECT * FROM {self.product} LIMIT {int(limit)}',
                           self.client)

    @staticmethod
    def _init_source(source, kwargs):
        return SQLScriptSource(source, **kwargs)


class SQLDump(io.FileLoaderMixin, ClientMixin, Task):
    """Dumps data from a SQL SELECT statement to a file(s)

    Parameters
    ----------
    source: str or pathlib.Path
        SQL script source, if str, the content is interpreted as the actual
        script, if pathlib.Path, the content of the file is loaded
    product: ploomber.products.product
        Product generated upon successful execution
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    client: ploomber.clients.{SQLAlchemyClient, DBAPIClient}, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]
    params: dict, optional
        Parameters to pass to the script, by default, the callable will
        be executed with a "product" (which will contain the product object).
        It will also include a "upstream" parameter if the task has upstream
        dependencies along with any parameters declared here. The source
        code is converted to a jinja2.Template for passing parameters,
        refer to jinja2 documentation for details
    chunksize: int, optional
        Number of rows per file, otherwise download the entire dataset in
        a single one. If not None, the product becomes a directory
    io_handler: ploomber.io.CSVIO or ploomber.io.ParquetIO, optional
        io handler to use (which controls the output format), currently
        only csv and parquet are supported. If None, it tries to infer the
        handler from the product's extension if that doesn't work, it uses
        io.CSVIO

    Notes
    -----
    The chunksize parameter is also set in cursor.arraysize object, this
    parameter can greatly speed up the dump for some databases when the
    driver uses cursors.arraysize as the number of rows to fetch on a single
    network trip, but this is driver-dependent, not all drivers implement
    this (cx_Oracle does it)
    """
    PRODUCT_CLASSES_ALLOWED = (File, GenericProduct)

    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 client=None,
                 params=None,
                 chunksize=10000,
                 io_handler=None):
        params = params or {}

        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)

        self._client = client
        self.chunksize = chunksize

        if io_handler is None:
            if self.product._identifier._raw.endswith('.parquet'):
                self.io_handler = io.ParquetIO
            else:
                self.io_handler = io.CSVIO
        else:
            self.io_handler = io_handler

    @staticmethod
    def _init_source(source, kwargs):
        return SQLQuerySource(source, **kwargs)

    def run(self):

        # render runtime parameters
        template = Template(str(self.source),
                            variable_start_string='[[',
                            variable_end_string=']]')
        _add_globals(template.environment)
        source_code = template.render(upstream=self.params.get('upstream'))

        path = Path(str(self.params['product']))
        handler = self.io_handler(path, chunked=bool(self.chunksize))

        self._logger.debug('Code: %s', source_code)

        cursor = self.client.connection.cursor()

        try:
            cursor.execute(source_code)
        except Exception as e:
            raise SQLTaskBuildError(type(self), source_code, e) from e

        if self.chunksize:
            i = 1
            headers = None
            cursor.arraysize = self.chunksize

            while True:
                self._logger.info('Fetching chunk {}...'.format(i))
                data = cursor.fetchmany()
                self._logger.info('Fetched chunk {}'.format(i))

                if i == 1:
                    headers = [c[0] for c in cursor.description]

                if not data:
                    break

                handler.write(data, headers)

                i = i + 1
        else:
            data = cursor.fetchall()
            headers = [c[0] for c in cursor.description]
            handler.write(data, headers)

        cursor.close()


# FIXME: this can be a lot faster for clients that transfer chunksize
# rows over the network
class SQLTransfer(ClientMixin, Task):
    """
    Transfers data from a SQL database to another (Note: this relies on
    pandas, only use it for small to medium size datasets)

    Parameters
    ----------
    source: str or pathlib.Path
        SQL script source, if str, the content is interpreted as the actual
        script, if pathlib.Path, the content of the file is loaded
    product: ploomber.products.product
        Product generated upon successful execution. For SQLTransfer, usually
        product.client != task.client. task.client represents the data source
        while product.client represents the data destination
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    client: ploomber.clients.SQLAlchemyClient, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]
    params: dict, optional
        Parameters to pass to the script, by default, the callable will
        be executed with a "product" (which will contain the product object).
        It will also include a "upstream" parameter if the task has upstream
        dependencies along with any parameters declared here. The source
        code is converted to a jinja2.Template for passing parameters,
        refer to jinja2 documentation for details
    chunksize: int, optional
        Number of rows to transfer on every chunk

    Notes
    ----
    This task is *not* intended to move large datasets, but a
    convenience way of transfering small to medium size datasets. It relies
    on pandas to read and write, which introduces a considerable overhead.
    """
    PRODUCT_CLASSES_ALLOWED = (PostgresRelation, SQLiteRelation,
                               GenericSQLRelation)

    @requires(['pandas'], 'SQLTransfer')
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 client=None,
                 params=None,
                 chunksize=10000):
        params = params or {}
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)
        self._client = client
        self.chunksize = chunksize

    @staticmethod
    def _init_source(source, kwargs):
        # TODO: this shoule be a FileSource
        return SQLQuerySource(source, **kwargs)

    def run(self):
        import pandas as pd

        source_code = str(self.source)
        product = self.params['product']

        # read from source_code, use connection from the Task
        self._logger.info('Fetching data...')
        dfs = pd.read_sql_query(source_code,
                                self.client.engine,
                                chunksize=self.chunksize)
        self._logger.info('Done fetching data...')

        for i, df in enumerate(dfs):
            self._logger.info('Storing chunk {i}...'.format(i=i))
            df.to_sql(name=product.name,
                      con=product.client.engine,
                      schema=product.schema,
                      if_exists='replace' if i == 0 else 'append',
                      index=False)


class SQLUpload(ClientMixin, Task):
    """
    Upload data to a SQL database from a parquet or a csv file. Note: this
    task relies uses pandas.to_sql which introduces some overhead. Only use it
    for small to medium size datasets. Each database usually come with a tool
    to upload data efficiently. If you are using PostgreSQL, check out the
    PostgresCopyFrom task.

    Parameters
    ----------
    source : str or pathlib.Path
        Path to parquet or a csv file to upload

    product : ploomber.products.product
        Product generated upon successful execution. The client for the product
        must be in the target database, where as task.client should be a client
        in the source database.

    dag : ploomber.DAG
        A DAG to add this task to

    name : str
        A str to indentify this task. Should not already exist in the dag

    client: ploomber.clients.SQLAlchemyClient, optional
        The client used to connect to the database and where the data will be
        uploaded. Only required
        if no dag-level client has been declared using dag.clients[class]

    params : dict, optional
        Parameters to pass to the script, by default, the callable will
        be executed with a "product" (which will contain the product object).
        It will also include a "upstream" parameter if the task has upstream
        dependencies along with any parameters declared here. The source
        code is converted to a jinja2.Template for passing parameters,
        refer to jinja2 documentation for details

    chunksize : int, optional
        Number of rows to transfer on every chunk

    io_handler : callable, optional
        A Python callable to read the source file,
        if None, it will tried to be inferred from the source file extension

    to_sql_kwargs : dict, optional
        Keyword arguments passed to the pandas.DataFrame.to_sql function,
        one useful parameter is "if_exists", which determines if the inserted
        rows should replace the table or just be appended

    Notes
    -----
    This task is *not* intended to move large datasets, but a
    convenience way of transfering small to medium size datasets. It relies
    on pandas to read and write, which introduces a considerable overhead.
    """
    PRODUCT_CLASSES_ALLOWED = (PostgresRelation, SQLiteRelation,
                               GenericSQLRelation)

    @requires(['pandas'], 'SQLUpload')
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 client=None,
                 params=None,
                 chunksize=None,
                 io_handler=None,
                 to_sql_kwargs=None):
        params = params or {}
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)
        self._client = client
        self.chunksize = chunksize
        self.io_handler = io_handler
        self.to_sql_kwargs = to_sql_kwargs or {}

    @staticmethod
    def _init_source(source, kwargs):
        return FileSource(str(source), **kwargs)

    def run(self):
        import pandas as pd

        product = self.params['product']
        path = str(self.source)

        mapping = {
            '.csv': pd.read_csv,
            '.parquet': pd.read_parquet,
        }

        if self.io_handler is None:
            extension = Path(path).suffix
            read_fn = mapping.get(extension)

            if not read_fn:
                raise ValueError(
                    'Could not infer reading function for '
                    'file with extension: {}'.format(extension),
                    'pass the function directly in the '
                    'io_handler argument')
        else:
            read_fn = self.io_handler

        self._logger.info('Reading data...')
        df = read_fn(path)
        self._logger.info('Done reading data...')

        df.to_sql(name=product.name,
                  con=self.client.engine,
                  schema=product.schema,
                  **self.to_sql_kwargs)


# TODO: provide more flexibility to configure the COPY statement
class PostgresCopyFrom(ClientMixin, Task):
    """Efficiently copy data to a postgres database using COPY FROM (faster
    alternative to SQLUpload for postgres). If using SQLAlchemy client
    for postgres is psycopg2. Replaces the table if exists.

    Parameters
    ----------
    source: str or pathlib.Path
        Path to parquet file to upload

    client: ploomber.clients.SQLAlchemyClient, optional
        The client used to connect to the database and where the data will be
        uploaded. Only required
        if no dag-level client has been declared using dag.clients[class]


    Notes
    -----
    Although this task does not depend on pandas for data i/o, it still
    needs it to dynamically create the table, after the table is created
    the COPY statement is used to upload the data
    """
    PRODUCT_CLASSES_ALLOWED = (PostgresRelation, )

    @requires(['pandas', 'psycopg2'], 'PostgresCopyFrom')
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 client=None,
                 params=None,
                 columns=None):
        params = params or {}
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)
        self._client = client
        self.columns = columns

    @staticmethod
    def _init_source(source, kwargs):
        return FileSource(str(source), **kwargs)

    def run(self):
        import pandas as pd

        product = self.params['product']
        df = pd.read_parquet(str(self.source))

        # create the table
        self._logger.info('Creating table...')
        df.head(0).to_sql(name=product.name,
                          con=self.client.engine,
                          schema=product.schema,
                          if_exists='replace',
                          index=False)
        self._logger.info('Done creating table.')

        # create file-like object
        f = StringIO()
        df.to_csv(f, sep='\t', na_rep='\\N', header=False, index=False)
        f.seek(0)

        # upload using copy
        cur = self.client.connection.cursor()

        self._logger.info('Copying data...')
        cur.copy_expert(f'COPY {product} FROM STDIN', f)

        f.close()
        cur.close()



from collections.abc import Mapping, MutableSequence
import keyword
import logging

import yaml


class FrozenJSON(object):
    """
    A facade for navigating a JSON-like object using attribute notation.
    Based on FrozenJSON from 'Fluent Python'
    """
    @classmethod
    def from_yaml(cls, path_to_file, *args, **kwargs):
        # load config file
        with open(path_to_file) as f:
            mapping = yaml.safe_load(f)
            # yaml.safe_load returns none if file is empty, but an empty
            # env.yaml is still useful for env.path
            mapping = mapping if mapping is not None else {}

        obj = cls(mapping, *args, **kwargs)

        # save path for reference, helps debugging
        obj._path_to_file = path_to_file

        logger = logging.getLogger(__name__)
        logger.debug('Loaded from file: {}'.format(obj._path_to_file))

        return obj

    def __new__(cls, arg):
        if isinstance(arg, Mapping):
            return super(FrozenJSON, cls).__new__(cls)

        elif isinstance(arg, MutableSequence):
            return [cls(item) for item in arg]
        else:
            return arg

    def __init__(self, mapping):
        self._logger = logging.getLogger(__name__)
        self._logger.debug('Loaded with params: {}'.format(mapping))
        self._path_to_file = None

        self._data = {}

        for key, value in mapping.items():
            if keyword.iskeyword(key):
                key += '_'

            self._data[key] = value

    def __getattr__(self, name):
        if hasattr(self._data, name):
            return getattr(self._data, name)
        else:
            return FrozenJSON(self._data[name])

    def __dir__(self):
        return self._data.keys()

    def __getitem__(self, key):
        value = self._data.get(key)

        if value is None:
            key_ = key if not isinstance(key, str) else "'%s'" % key
            msg = ('Key error: {}, available keys are: {}'.format(
                key_, self._data.keys()))
            if self._path_to_file is not None:
                msg += '. File loaded from {}'.format(self._path_to_file)
            raise KeyError(msg)

        return value

    def __str__(self):
        if self._path_to_file:
            return self._path_to_file
        else:
            return str(self._data)

    def __repr__(self):
        return 'FrozenJSON({})'.format(str(self))



"""
Implementation of:

$ plomber install

This command runs a bunch of pip/conda commands (depending on what's available)
and it does the *right thing*: creating a new environment if needed, and
locking dependencies.
"""
import sys
import json
import os
import shutil
from pathlib import Path
from contextlib import contextmanager

import click
import yaml

from ploomber.io._commander import Commander
from ploomber.exceptions import BaseException
from ploomber.util.util import check_mixed_envs
from ploomber.cli.io import command_endpoint
from ploomber.telemetry import telemetry
from ploomber.util._sys import _python_bin

_SETUP_PY = 'setup.py'

_REQS_LOCK_TXT = 'requirements.lock.txt'
_REQS_TXT = 'requirements.txt'

_ENV_YML = 'environment.yml'
_ENV_LOCK_YML = 'environment.lock.yml'

_PYTHON_BIN_NAME = _python_bin()


@command_endpoint
@telemetry.log_call('install')
def main(use_lock, create_env=None, use_venv=False):
    """
    Install project, automatically detecting if it's a conda-based or pip-based
    project.

    Parameters
    ---------
    use_lock : bool
        If True Uses requirements.lock.txt/environment.lock.yml and
        requirements.dev.lock.txt/environment.dev.lock.yml files. If False
        uses regular files and creates the lock ones after installing
        dependencies. If None, it uses lock files if they exist, if they don't
        it uses regular files

    create_env : bool, default=None
        If True, creates a new environment, if False, it installs in the
        current environment. If None, it creates a new environment if there
        isn't one already active

    use_venv : bool, default=False
        Force to use Python's venv module, ignoring conda if installed
    """
    USE_CONDA = shutil.which('conda') and not use_venv
    ENV_YML_EXISTS = Path(_ENV_YML).exists()
    ENV_LOCK_YML_EXISTS = Path(_ENV_LOCK_YML).exists()
    REQS_TXT_EXISTS = Path(_REQS_TXT).exists()
    REQS_LOCK_TXT_EXISTS = Path(_REQS_LOCK_TXT).exists()

    if use_lock is None:
        if USE_CONDA:
            use_lock = ENV_LOCK_YML_EXISTS
        else:
            use_lock = REQS_LOCK_TXT_EXISTS

    if use_lock and not ENV_LOCK_YML_EXISTS and not REQS_LOCK_TXT_EXISTS:
        raise BaseException(
            "Expected an environment.lock.yaml "
            "(conda) or requirements.lock.txt (pip) in the current "
            "directory. Add one of them and try again.",
            type_='no_lock')
    elif not use_lock and not ENV_YML_EXISTS and not REQS_TXT_EXISTS:
        raise BaseException(
            "Expected an environment.yaml (conda)"
            " or requirements.txt (pip) in the current directory."
            " Add one of them and try again.",
            type_='no_env_requirements')
    elif (not USE_CONDA and use_lock and ENV_LOCK_YML_EXISTS
          and not REQS_LOCK_TXT_EXISTS):
        raise BaseException(
            "Found env environment.lock.yaml "
            "but conda is not installed. Install conda or add a "
            "requirements.lock.txt to use pip instead",
            type_='no_conda')
    elif (not USE_CONDA and not use_lock and ENV_YML_EXISTS
          and not REQS_TXT_EXISTS):
        raise BaseException(
            "Found environment.yaml but conda is not installed."
            " Install conda or add a requirements.txt to use pip instead",
            type_='no_conda2')
    elif USE_CONDA and use_lock and ENV_LOCK_YML_EXISTS:
        # TODO: emit warnings if unused requirements.txt?
        main_conda(use_lock=True,
                   create_env=create_env
                   if create_env is not None else _should_create_conda_env())
    elif USE_CONDA and not use_lock and ENV_YML_EXISTS:
        # TODO: emit warnings if unused requirements.txt?
        main_conda(use_lock=False,
                   create_env=create_env
                   if create_env is not None else _should_create_conda_env())
    else:
        # TODO: emit warnings if unused environment.yml?
        main_pip(use_lock=use_lock,
                 create_env=create_env
                 if create_env is not None else not telemetry.in_virtualenv())


def main_pip(use_lock, create_env=True):
    """
    Install pip-based project (uses venv), looks for requirements.txt files

    Parameters
    ----------
    start_time : datetime
        The initial runtime of the function.

    use_lock : bool
        If True Uses requirements.txt and requirements.dev.lock.txt files

    create_env : bool
        If True, it uses the venv module to create a new virtual environment,
        then installs the dependencies, otherwise it installs the dependencies
        in the current environment
    """
    reqs_txt = _REQS_LOCK_TXT if use_lock else _REQS_TXT
    reqs_dev_txt = ('requirements.dev.lock.txt'
                    if use_lock else 'requirements.dev.txt')

    cmdr = Commander()

    # TODO: modify readme to add how to activate env? probably also in conda
    name = Path('.').resolve().name

    try:
        _run_pip_commands(cmdr, create_env, name, reqs_dev_txt, reqs_txt,
                          use_lock)
    except Exception as e:
        cmd = f'pip install --requirement {reqs_txt}'
        raise BaseException('Failed to setup your environment. '
                            f'Invoke pip manually.\n{cmd}\n\n') from e


def _run_pip_commands(cmdr, create_env, name, reqs_dev_txt, reqs_txt,
                      use_lock):
    if create_env:
        venv_dir = f'venv-{name}'
        cmdr.print('Creating venv...')
        cmdr.run(_PYTHON_BIN_NAME,
                 '-m',
                 'venv',
                 venv_dir,
                 description='Creating venv')

        # add venv_dir to .gitignore if it doesn't exist
        if Path('.gitignore').exists():
            with open('.gitignore') as f:
                if venv_dir not in f.read():
                    cmdr.append_inline(venv_dir, '.gitignore')
        else:
            cmdr.append_inline(venv_dir, '.gitignore')

        folder, bin_name = _get_pip_folder_and_bin_name()
        pip = str(Path(venv_dir, folder, bin_name))

        if os.name == 'nt':
            cmd_activate = f'{venv_dir}\\Scripts\\Activate.ps1'
        else:
            cmd_activate = f'source {venv_dir}/bin/activate'
    else:
        cmdr.print('Installing in current venv...')
        pip = 'pip'
        cmd_activate = None

    # FIXME: using an old version of pip may lead to broken environments, so
    # we need to ensure we upgrade before installing dependencies.

    if Path(_SETUP_PY).exists():
        _pip_install_setup_py_pip(cmdr, pip)

    _pip_install(cmdr, pip, lock=not use_lock, requirements=reqs_txt)

    if Path(reqs_dev_txt).exists():
        _pip_install(cmdr, pip, lock=not use_lock, requirements=reqs_dev_txt)

    _next_steps(cmdr, cmd_activate)


def main_conda(use_lock, create_env=True):
    """
    Install conda-based project, looks for environment.yml files

    Parameters
    ----------
    use_lock : bool
        If True Uses environment.lock.yml and environment.dev.lock.yml files


    create_env : bool
        If True, it uses the venv module to create a new virtual environment,
        then installs the dependencies, otherwise it installs the dependencies
        in the current environment
    """
    env_yml = _ENV_LOCK_YML if use_lock else _ENV_YML

    # TODO: ensure ploomber-scaffold includes dependency file (including
    # lock files in MANIFEST.in
    cmdr = Commander()

    # TODO: provide helpful error messages on each command

    if create_env:
        with open(env_yml) as f:
            env_name = yaml.safe_load(f)['name']

        current_env = _current_conda_env_name()

        if env_name == current_env:
            err = (f'{env_yml} will create an environment '
                   f'named {env_name!r}, which is the current active '
                   'environment. Activate a different one and try '
                   'again: conda activate base')
            telemetry.log_api("install-error",
                              metadata={
                                  'type': 'env_running_conflict',
                                  'exception': err
                              })
            raise BaseException(err)
    else:
        env_name = _current_conda_env_name()

    # get current installed envs
    conda = shutil.which('conda')
    mamba = shutil.which('mamba')

    # if already installed and running on windows, ask to delete first,
    # otherwise it might lead to an intermittent error (permission denied
    # on vcruntime140.dll)
    if os.name == 'nt' and create_env:
        envs = cmdr.run(conda, 'env', 'list', '--json', capture_output=True)
        already_installed = any([
            env for env in json.loads(envs)['envs']
            # only check in the envs folder, ignore envs in other locations
            if 'envs' in env and env_name in env
        ])

        if already_installed:
            err = (f'Environment {env_name!r} already exists, '
                   f'delete it and try again '
                   f'(conda env remove --name {env_name})')
            telemetry.log_api("install-error",
                              metadata={
                                  'type': 'duplicate_env',
                                  'exception': err
                              })
            raise BaseException(err)

    pkg_manager = mamba if mamba else conda

    try:
        _run_conda_commands(cmdr, pkg_manager, create_env, env_yml, env_name,
                            use_lock, conda)
    except Exception as e:
        if create_env:
            cmd = f'conda env create --file {env_yml} --force'
        else:
            cmd = f'conda env update --file {env_yml} --name {env_name}'
        raise BaseException('Failed to setup your environment. '
                            f'Invoke conda manually.\n{cmd}\n\n') from e


def _run_conda_commands(
    cmdr,
    pkg_manager,
    create_env,
    env_yml,
    env_name,
    use_lock,
    conda,
):
    if create_env:
        cmdr.print('Creating conda env...')
        cmdr.run(pkg_manager,
                 'env',
                 'create',
                 '--file',
                 env_yml,
                 '--force',
                 description='Creating env')
    else:
        cmdr.print('Installing in current conda env...')

        cmdr.run(pkg_manager,
                 'env',
                 'update',
                 '--file',
                 env_yml,
                 '--name',
                 env_name,
                 description='Installing dependencies')

    if Path(_SETUP_PY).exists():
        _pip_install_setup_py_conda(cmdr, env_name)

    if not use_lock:
        env_lock = cmdr.run(conda,
                            'env',
                            'export',
                            '--no-build',
                            '--name',
                            env_name,
                            description='Locking dependencies',
                            capture_output=True)
        Path(_ENV_LOCK_YML).write_text(env_lock)

    _try_conda_install_and_lock_dev(cmdr,
                                    pkg_manager,
                                    env_name,
                                    use_lock=use_lock)

    cmd_activate = (f'conda activate {env_name}' if create_env else None)
    _next_steps(cmdr, cmd_activate)


def _should_create_conda_env():
    # not in conda env or running in base conda env
    return (not telemetry.is_conda()
            or (telemetry.is_conda() and _current_conda_env_name() == 'base'))


def _current_conda_env_name():
    return (os.environ.get('CONDA_DEFAULT_ENV')
            or Path(sys.executable).parents[1].name)


def _get_pip_folder_and_bin_name():
    folder = 'Scripts' if os.name == 'nt' else 'bin'
    bin_name = 'pip.exe' if os.name == 'nt' else 'pip'
    return folder, bin_name


def _find_conda_root(conda_bin):
    conda_bin = Path(conda_bin)

    for parent in conda_bin.parents:
        # I've seen variations of this. on windows: Miniconda3 and miniconda3
        # on linux miniconda3, anaconda and miniconda
        if parent.name.lower() in {'miniconda3', 'miniconda', 'anaconda3'}:
            return parent
    err = ('Failed to locate conda root from '
           f'directory: {str(conda_bin)!r}. Please submit an issue: '
           'https://github.com/ploomber/ploomber/issues/new')
    telemetry.log_api("install-error",
                      metadata={
                          'type': 'no_conda_root',
                          'exception': err
                      })
    raise BaseException(err)


def _path_to_pip_in_env_with_name(conda_bin, env_name):
    conda_root = _find_conda_root(conda_bin)
    folder, bin_name = _get_pip_folder_and_bin_name()
    return str(conda_root / 'envs' / env_name / folder / bin_name)


def _locate_pip_inside_conda(env_name):
    """
    Locates pip inside the conda env with a given name
    """
    pip = _path_to_pip_in_env_with_name(shutil.which('conda'), env_name)

    # this might happen if the environment does not contain python/pip
    if not Path(pip).exists():
        err = (f'Could not locate pip in environment {env_name!r}, make sure '
               'it is included in your environment.yml and try again')
        telemetry.log_api("install-error",
                          metadata={
                              'type': 'no_pip_env',
                              'exception': err
                          })
        raise BaseException(err)

    return pip


def _pip_install_setup_py_conda(cmdr, env_name):
    """
    Call "pip install --editable ." if setup.py exists. Automatically locates
    the appropriate pip binary inside the conda env given the env name
    """
    pip = _locate_pip_inside_conda(env_name)
    _pip_install_setup_py_pip(cmdr, pip)


def _pip_install_setup_py_pip(cmdr, pip):
    cmdr.run(pip,
             'install',
             '--editable',
             '.',
             description='Installing project')


def _try_conda_install_and_lock_dev(cmdr, pkg_manager, env_name, use_lock):
    env_yml = 'environment.dev.lock.yml' if use_lock else 'environment.dev.yml'

    if Path(env_yml).exists():
        cmdr.run(pkg_manager,
                 'env',
                 'update',
                 '--file',
                 env_yml,
                 '--name',
                 env_name,
                 description='Installing dev dependencies')

        if not use_lock:
            env_lock = cmdr.run(shutil.which('conda'),
                                'env',
                                'export',
                                '--no-build',
                                '--name',
                                env_name,
                                description='Locking dev dependencies',
                                capture_output=True)
            Path('environment.dev.lock.yml').write_text(env_lock)


def _next_steps(cmdr, cmd_activate):
    cmdr.success('Next steps')
    message = f'$ {cmd_activate}\n' if cmd_activate else ''
    cmdr.print((f'{message}$ ploomber build'))
    cmdr.success()


def _pip_install(cmdr, pip, lock, requirements=_REQS_TXT):
    """Install and freeze requirements

    Parameters
    ----------
    cmdr
        Commander instance

    pip
        Path to pip binary

    lock
        If true, locks dependencies and stores them in a requirements.lock.txt
    """
    cmdr.run(pip,
             'install',
             '--requirement',
             requirements,
             description='Installing dependencies')

    if lock:
        pip_lock = cmdr.run(pip,
                            'freeze',
                            '--exclude-editable',
                            description='Locking dependencies',
                            capture_output=True)
        check_mixed_envs(pip_lock)
        name = Path(requirements).stem
        Path(f'{name}.lock.txt').write_text(pip_lock)


def _environment_yml_has_python(path):
    with open(path) as f:
        env_yml = yaml.safe_load(f)

    deps = env_yml.get('dependencies', [])

    has_python = False
    idx = None

    for i, line in enumerate(deps):
        if isinstance(line, str) and line.startswith('python'):
            has_python = True
            idx = i
            break

    if has_python:
        env_yml['dependencies'].pop(idx)

    return has_python, env_yml


@contextmanager
def check_environment_yaml(path, enable=True):
    has_python, env_yml = _environment_yml_has_python(path)
    TMP_FILENAME = '.ploomber-conda-tmp.yml'

    if has_python and enable:
        path_to_use = Path(TMP_FILENAME)
        path_to_use.write_text(yaml.dump(env_yml))
        click.secho(
            f'WARNING: {path!r} contains Python as '
            'dependency, ignoring it as it may break '
            'the current environment',
            fg='yellow')
    else:
        path_to_use = Path(path)

    try:
        yield str(path_to_use)
    finally:
        if Path(TMP_FILENAME).exists():
            path_to_use.unlink()



from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import html
import mistune


class HighlightRenderer(mistune.Renderer):
    """mistune renderer with syntax highlighting

    Notes
    -----
    Source: https://github.com/lepture/mistune#renderer
    """
    def block_code(self, code, lang):
        if not lang:
            return '\n<pre><code>%s</code></pre>\n' % \
                mistune.escape(code)
        lexer = get_lexer_by_name(lang, stripall=True)
        formatter = html.HtmlFormatter()
        return highlight(code, lexer, formatter)



"""
Handling tasks that generate multiple products
"""
from reprlib import Repr
from collections.abc import Mapping

from ploomber.products.metadata import MetadataCollection
from ploomber.constants import TaskStatus


# TODO: raise error (or warning?) if products[*].client are different
# objects - that should not be possible
# NOTE: what do we do if metadata among products does not match?
# what if there are some products are File and others are not?
# maybe do not allow that scenario?
class ProductsContainer:
    """
    An iterator that when initialized with a sequence behaves just like it
    but when initialized with a mapping, iterates over values (instead of
    keys), this non-standard behavior but it is needed to simpplify the
    MetaProduct API since it has to work with lists and dictionaries
    """
    def __init__(self, products):
        self.products = products

        if isinstance(self.products, Mapping):
            self._first = list(self.products.values())[0]
        else:
            self._first = self.products[0]

    @property
    def first(self):
        return self._first

    def __iter__(self):
        if isinstance(self.products, Mapping):
            for product in self.products.values():
                yield product
        else:
            for product in self.products:
                yield product

    def __getitem__(self, key):
        return self.products[key]

    def to_json_serializable(self):
        """Returns a JSON serializable version of this product
        """
        if isinstance(self.products, Mapping):
            return {
                name: str(product)
                for name, product in self.products.items()
            }
        else:
            return list(str(product) for product in self.products)

    def __len__(self):
        return len(self.products)

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, str(self.products))

    def __str__(self):
        return str(self.products)


class ClientContainer:
    def __init__(self, products):
        self._products = products

    def close(self):
        for product in self._products:
            if product.client:
                product.client.close()

    def download(self, local, destination=None):
        client = self._products.first.client
        return client.download(local, destination)


# NOTE: rename this to ProductCollection?
# TODO: create a superclass of Product, define the interface there,
# then make MetaProduct and Product inherit from it
class MetaProduct(Mapping):
    """
    Exposes a Product-like API to allow Tasks to create more than one Product,
    it is automatically instantiated when a Task is initialized with a
    sequence or a mapping object in the product parameter. While it is
    recommended for Tasks to only have one Product (to keep them simple),
    in some cases it makes sense. For example, a Jupyter notebook
    (executed via NotebookRunner), for fitting a model might as well serialize
    the things such as the model and any data preprocessors
    """
    def __init__(self, products):
        container = ProductsContainer(products)

        self.products = container
        self.metadata = MetadataCollection(container)
        self.clients = ClientContainer(container)

        self._repr = Repr()

    @property
    def task(self):
        # TODO: validate same task
        return self.products[0].task

    @property
    def client(self):
        return self.clients

    @task.setter
    def task(self, value):
        for p in self.products:
            try:
                p.task = value
            except AttributeError as e:
                raise AttributeError(
                    "Expected MetaProduct to initialize with Product "
                    "instancess (which have a 'task' attribute), but "
                    f"got {p!r}, an object of type {type(p)}. Replace it "
                    "with a valid Product object. If this is a file, use "
                    f"File({p!r})") from e

    def exists(self):
        return all([p.exists() for p in self.products])

    def missing(self):
        # used to throw a meaninful error message when some products do not
        # exist after executing the user's code
        return [p for p in self.products if not p.exists()]

    def delete(self, force=False):
        for product in self.products:
            product.delete(force)

    def download(self):
        for product in self.products:
            product.download()

    def upload(self):
        for product in self.products:
            product.upload()

    def _is_outdated(self, outdated_by_code=True):
        is_outdated = [
            p._is_outdated(outdated_by_code=outdated_by_code)
            for p in self.products
        ]

        if set(is_outdated) == {False}:
            return False

        if set(is_outdated) <= {TaskStatus.WaitingDownload, False}:
            return TaskStatus.WaitingDownload

        return any(is_outdated)

    def _outdated_data_dependencies(self):
        return any([p._outdated_data_dependencies() for p in self.products])

    def _outdated_code_dependency(self):
        return any([p._outdated_code_dependency() for p in self.products])

    def to_json_serializable(self):
        """Returns a JSON serializable version of this product
        """
        # NOTE: this is used in tasks where only JSON serializable parameters
        # are supported such as NotebookRunner that depends on papermill
        return self.products.to_json_serializable()

    def render(self, params, **kwargs):
        for p in self.products:
            p.render(params, **kwargs)

    def __repr__(self):
        content = self._repr.repr1(self.products.products, level=2)
        return f'{type(self).__name__}({content})'

    def __str__(self):
        return str(self.products)

    def __iter__(self):
        for product in self.products:
            yield product

    def __getitem__(self, key):
        return self.products[key]

    def __len__(self):
        return len(self.products)

    @property
    def _remote(self):
        return self.products.first._remote

    def _is_remote_outdated(self, outdated_by_code):
        return self.products.first._is_remote_outdated(outdated_by_code)



from ploomber.exceptions import MissingClientError
from ploomber.util.dotted_path import DottedPath


class ClientMixin:
    """
    A mixin that exposes a client property. Looks at the _client attribute,
    then looks at self.dag. If _client returns a DottedSpecPath, it is called,
    replaced, and returned

    Raises
    ------
    ValueError
        If there is no valid client to use
    """
    @property
    def client(self):
        if self._client is None:
            dag_client = self.dag.clients.get(type(self))

            if dag_client is None:
                raise MissingClientError(
                    f'{type(self).__name__} must be initialized with a '
                    'client. Pass a client directly or set a DAG-level one')

            return dag_client

        if isinstance(self._client, DottedPath):
            # NOTE: should we test the output ot self._client()?
            # maybe check if it's a subclass of the Client abstract class
            self._client = self._client()

        return self._client



"""
TaskGroup allows group tasks together. It enables syntactic sugar when adding
dependencies explicitly (e.g., (t1 + t2) >> t3). It also allows creating
multiple tasks at once using a parameters array.

There are two potential cases that we haven't implemented:
    1. Automatically generate an array of tasks to apply a transformation over
        file partitions. e.g., user supplies code that cannot be easily
        parallelized using numpy/pandas + the column(s) to use for
        partitioning. Then we partition in equal parts ensuring that all rows
        with the same id end up in the same partition, generate N tasks (one
        per partition) and output a final directory with the outputs of each
        task. Since parquet supports partitions, the parquet partitions can be
        used without having to join them.
    2. Same use case as before but this time, the tasks look like a single one
        from the DAGs perspective. Inside they are executed in a for loop. This
        can be useful to scale up batch computations in a single-machine.
        Often DS/ML practitioners run into memory errors because pandas/numpy
        have to load the entire dataset to memory, if we process the files
        in partitions, we can process the whole dataset in parts and avoid
        memory issues. The only pre-requisite for this would be for upstream
        dependencies to generate their product in parquet partitions.

These are nice-to-have features that require a considerable effort.

A few notes from a first implemenation attempt:
    * Although use cases (1) and (2) sound similar, from the dag perspective
        they are different (1) generates multiple taks, (2) generates a single
        task that looks like one from the outside. It might be better to
        have separate implementations, instead of a single TaskGroup
"""
from collections.abc import Mapping
from pathlib import Path
from copy import deepcopy, copy

from jinja2 import Template

from ploomber.products import File
from ploomber.util import isiterable, ParamGrid
from ploomber.products.mixins import SQLProductMixin


class TaskGroup:
    """
    A collection of Tasks, used internally for enabling operador overloading

    (task1 + task2) >> task3
    """
    def __init__(self, tasks):
        self.tasks = tasks

    def __iter__(self):
        for t in self.tasks:
            yield t

    def __len__(self):
        return len(self.tasks)

    def __add__(self, other):
        if isiterable(other):
            return TaskGroup(list(self.tasks) + list(other))
        else:
            return TaskGroup(list(self.tasks) + [other])

    def __radd__(self, other):
        if isiterable(other):
            return TaskGroup(list(other) + list(self.tasks))
        else:
            return TaskGroup([other] + list(self.tasks))

    def set_upstream(self, other):
        if isiterable(other):
            for t in self.tasks:
                for o in other:
                    t.set_upstream(other)
        else:
            for t in self.tasks:
                t.set_upstream(other)

    def __rshift__(self, other):
        other.set_upstream(self)
        # return other so a >> b >> c works
        return other

    @classmethod
    def from_params(cls,
                    task_class,
                    product_class,
                    product_primitive,
                    task_kwargs,
                    dag,
                    params_array,
                    name=None,
                    namer=None,
                    resolve_relative_to=None,
                    on_render=None,
                    on_finish=None,
                    on_failure=None):
        """
        Build a group of tasks of the same class from an array of parameters
        using the same source. Generates one task per element in params_array.

        Parameters
        ----------
        task_class : class
            The task class for all generated tasks

        product_class : class
            The class used to initialize the products

        product_primitive : str or list
            The object used to initialize the product. For File, this is a str,
            for SQL-like products, it must be a list.

        task_kwargs : dict
            Task keyword arguments passed to the constructor, must not contain:
            dag, name, params, nor product, as this parameters are passed by
            this function. Must include source.

        dag : ploomber.DAG
            The DAG object to add these tasks to

        params_array : list
            Each element is passed to the "params" argument on each task. This
            determines the number of tasks to be created. "name" is added to
            each
            element.

        name : str, default=None
            The name prefix for each of the tasks in the task group. If namer
            is None, this must not be None.

        namer : callable, default=None
            A function that receives a single argument (the task parameters
            dict) and returns a string used as a task name. If name is None,
            this must not be None.

        resolve_relative_to : str or pathlib.Path, default=None
            If not None, paths in File products are resolved to be absolute
            paths

        """
        if name is None and namer is None:
            raise ValueError(
                'Only one of name and namer can be None, but not both')

        # validate task_kwargs
        if 'dag' in task_kwargs:
            raise KeyError('dag should not be part of task_kwargs')

        if 'name' in task_kwargs:
            raise KeyError('name should not be part of task_kwargs')

        if 'params' in task_kwargs:
            raise KeyError('params should not be part of task_kwargs')

        if 'product' in task_kwargs:
            raise KeyError('product should not be part of task_kwargs')

        if 'source' not in task_kwargs:
            raise KeyError('source should be in task_kwargs')

        # TODO: validate {{index}} appears in product - maybe all products
        # should have a way to extract which placeholders exist?

        tasks_all = []

        for index, params in enumerate(params_array):

            # each task should get a different deep copy, primarily cause they
            # should have a different product
            kwargs = deepcopy(task_kwargs)
            # params should also be different copies, otherwise if the same
            # grid is re-used in several tasks, modifying anything there will
            # have side-effects
            params = deepcopy(params)

            # user provided a namer function
            if namer:
                task_name = namer(params)
                # if function provided, do not use indexing
                index = None
            # no namer function, just add an index
            else:
                task_name = name + str(index)

            if isinstance(product_primitive, str):
                product_primitive_to_use = Template(product_primitive).render(
                    name=task_name)
            else:
                product_primitive_to_use = product_primitive

            # add index to product primitive
            if product_class is File or issubclass(product_class,
                                                   SQLProductMixin):
                product = _init_product(
                    product_class,
                    product_primitive_to_use,
                    index,
                    resolve_relative_to=resolve_relative_to,
                    params=params,
                )
            else:
                raise NotImplementedError('TaskGroup only sypported for '
                                          'File and SQL products. '
                                          f'{product_class} is not supported')

            t = task_class(product=product,
                           dag=dag,
                           name=task_name,
                           params=params,
                           **kwargs)

            if on_render:
                t.on_render = on_render

            if on_finish:
                t.on_finish = on_finish

            if on_failure:
                t.on_failure = on_failure

            tasks_all.append(t)

        return cls(tasks_all)

    @classmethod
    def from_grid(cls,
                  task_class,
                  product_class,
                  product_primitive,
                  task_kwargs,
                  dag,
                  grid,
                  name=None,
                  namer=None,
                  resolve_relative_to=None,
                  on_render=None,
                  on_finish=None,
                  on_failure=None,
                  params=None):
        """
        Build a group of tasks of the same class from an grid of parameters
        using the same source.

        Parameters
        ----------
        grid : dict or list of dicts
            If dict, all combinations of individual parameters are generated.
            If list of dicts, each dict is processed individually, then
            concatenated to generate the final set.

        params : dict
            Values that will remain constant

        Notes
        -----
        All parameters, except for grid are the same as in .from_params
        """
        params_array = ParamGrid(grid, params=params).product()
        return cls.from_params(task_class=task_class,
                               product_class=product_class,
                               product_primitive=product_primitive,
                               task_kwargs=task_kwargs,
                               dag=dag,
                               name=name,
                               params_array=params_array,
                               namer=namer,
                               resolve_relative_to=resolve_relative_to,
                               on_render=on_render,
                               on_finish=on_finish,
                               on_failure=on_failure)


def _init_product(product_class, product_primitive, index, resolve_relative_to,
                  params):
    if isinstance(product_primitive, Mapping):
        return {
            key: _init_product(product_class, primitive, index,
                               resolve_relative_to, params)
            for key, primitive in product_primitive.items()
        }
    elif isinstance(product_primitive, str):
        return _init_product_with_str(product_class, product_primitive, index,
                                      resolve_relative_to, params)
    # is there a better way to check this? Sequence also matches str/bytes
    elif isinstance(product_primitive, (list, tuple)):
        return _init_product_with_sql_elements(product_class,
                                               product_primitive, index,
                                               params)
    else:
        raise NotImplementedError('TaskGroup only supported for task dict '
                                  'and str product primitives. Got '
                                  f'{product_primitive}, an object of type '
                                  f'{type(product_primitive).__name__}')


def _init_product_with_str(product_class, product_primitive, index,
                           resolve_relative_to, params):
    if index is not None:
        path = (Path(product_primitive) if resolve_relative_to is None else
                Path(resolve_relative_to, product_primitive).resolve())

        suffix = ''.join(path.suffixes)
        filename = path.name.replace(suffix, '')
        filename_with_index = f'{filename}-{index}{suffix}'

        path_final = Template(
            str(path.parent / filename_with_index),
            variable_start_string='[[',
            variable_end_string=']]',
        ).render(**params)

        return product_class(path_final)
    else:
        return product_class(product_primitive)


def _init_product_with_sql_elements(product_class, product_primitive, index,
                                    params):
    # this could be [schema, name, type] or just [name, type]
    index_to_change = 1 if len(product_primitive) == 3 else 0
    updated = copy(product_primitive)

    table_name = product_primitive[index_to_change] + f'-{index}'

    table_name_final = Template(
        table_name,
        variable_start_string='[[',
        variable_end_string=']]',
    ).render(**params)

    updated[index_to_change] = table_name_final

    return product_class(updated)



import abc
from ploomber.io import TerminalWriter
from io import StringIO


class Message:
    def __init__(self, task, message, obj=None):
        self._task = task
        self._message = message
        self._obj = obj

    @property
    def header(self):
        return repr(self._task)

    @property
    def sub_header(self):
        loc = self._task.source.loc
        return None if not loc else str(loc)

    @property
    def message(self):
        return self._message

    @property
    def obj(self):
        return self._obj


class MessageCollector(abc.ABC):
    """Collect messages and display them as a single str with task names
    Utilities for exception handling

    When a DAG is rendered or built, exceptions/warnings might happen, this
    class helps collect them all to show a meaninful error message where each
    one is shown along with the task name it generated it.

    """
    def __init__(self, messages=None):
        self.messages = messages or []

    def append(self, task, message, obj=None):
        self.messages.append(Message(task=task, message=message, obj=obj))

    @abc.abstractmethod
    def __str__(self):
        pass

    def _to_str(self, name=None, file=None, writer_kwargs=None):
        """
        Return the string representation of the collected messages

        Parameters
        ----------
        name
            Title to show at the end
        file
            Text stream to use. If None, uses a temporary StringIO object
        writer_kwargs
            Extra keyword arguments passed to the terminal writer
        """
        writer_kwargs = writer_kwargs or dict()

        if file is None:
            sio = StringIO()
        else:
            sio = file

        sio.write('\n')

        self.tw = TerminalWriter(file=sio)

        if name:
            self.tw.sep('=', title=name, **writer_kwargs)
        else:
            self.tw.sep('=', **writer_kwargs)

        for msg in self.messages:
            self.tw.sep('-', title=msg.header, **writer_kwargs)

            sub_header = msg.sub_header

            if sub_header:
                self.tw.sep('-', title=sub_header, **writer_kwargs)

            self.tw._write_source(msg.message.splitlines(), lexer='pytb')

        n = len(self)
        t = 'task' if n == 1 else 'tasks'
        self.tw.sep('=', title=f'Summary ({n} {t})', **writer_kwargs)

        for msg in self.messages:
            # TODO: include original exception type and error message in
            # summary
            self.tw.write(f'{msg.header}\n')

        if name:
            self.tw.sep('=', title=name, **writer_kwargs)
        else:
            self.tw.sep('=', **writer_kwargs)

        sio.seek(0)
        out = sio.read()

        if file is None:
            sio.close()

        return out

    def __len__(self):
        return len(self.messages)

    def __bool__(self):
        return bool(self.messages)

    def __iter__(self):
        for message in self.messages:
            yield message


class BuildExceptionsCollector(MessageCollector):
    def __str__(self):
        return self._to_str(name='DAG build failed',
                            file=None,
                            writer_kwargs=dict(red=True))


class RenderExceptionsCollector(MessageCollector):
    def __str__(self):
        return self._to_str(name='DAG render failed',
                            file=None,
                            writer_kwargs=dict(red=True))


class BuildWarningsCollector(MessageCollector):
    def __str__(self):
        return self._to_str(name='DAG build with warnings',
                            file=None,
                            writer_kwargs=dict(yellow=True))


class RenderWarningsCollector(MessageCollector):
    def __str__(self):
        return self._to_str(name='DAG render with warnings',
                            file=None,
                            writer_kwargs=dict(yellow=True))
